# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, ClassVar

from brayns.network import Future, Instance, JsonRpcReply
from brayns.utils import JsonSchema, Vector3, deserialize_schema

from .model import Model, deserialize_model


@dataclass
class LoaderInfo:
    """Loader description.

    :param name: Loader name.
    :type name: str
    :param extensions: Supported extensions without the dot.
    :type extensions: list[str]
    :param schema: Parameters JSON schema (low level).
    :type schema: JsonSchema
    :param binary: Wether the loader can load binary data.
    :type binary: bool
    :param schema: Plugin required to have the loader available.
    :type schema: str
    """

    name: str
    extensions: list[str]
    schema: JsonSchema
    binary: bool = False
    plugin: str = ""


def get_loaders(instance: Instance) -> list[LoaderInfo]:
    """Retreive all available loaders from an instance.

    :param instance: Instance.
    :type instance: Instance
    :return: List of loader descriptions.
    :rtype: list[LoaderInfo]
    """
    result = instance.request("get-loaders")
    return [_deserialize_loader(loader) for loader in result]


class Loader(ABC):
    """Base class for all loaders.

    Loader are used to load models from files into an instance.

    Available loaders for a given instance depends on the loaded plugins.

    The list of available loaders and supported extensions for a given instance
    can be queried using ``get_loaders(instance)``.
    """

    @classmethod
    @property
    @abstractmethod
    def name(cls) -> str:
        """Get the loader name.

        Can be compared with ``LoaderInfo.name``.

        :return: Loader name.
        :rtype: str
        """
        pass

    @abstractmethod
    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        pass

    def load_models(self, instance: Instance, path: str) -> list[Model]:
        """Load the given file into an instance and return the models.

        :param instance: Instance.
        :type instance: Instance
        :param path: Model(s) file path.
        :type path: str
        :return: List of created models.
        :rtype: list[Model]
        """
        task = self.load_models_task(instance, path)
        return task.wait_for_result()

    def load_models_task(self, instance: Instance, path: str) -> Future[list[Model]]:
        """Asynchronous version of ``load_models``.

        :param instance: Instance.
        :type instance: Instance
        :param path: Model(s) file path.
        :type path: str
        :return: Future to monitor the task.
        :rtype: Future[list[Model]]
        """
        params = _get_params(self, path=path)
        task = instance.task("add-model", params)
        return Future(task, _get_models)

    def upload_models(
        self, instance: Instance, format: str, data: bytes
    ) -> list[Model]:
        """Load models from binary data.

        As the model format cannot be deduced from a path, it must be specified.

        The format is the file extension without the dot, check the class
        variables to see supported formats (ex: MeshLoader.PLY).

        :param instance: Instance.
        :type instance: Instance
        :param format: Model format (see loader class variables).
        :type format: str
        :param data: Model binary data.
        :type data: bytes
        :return: List of created models.
        :rtype: list[Model]
        """
        task = self.upload_models_task(instance, format, data)
        return task.wait_for_result()

    def upload_models_task(
        self, instance: Instance, format: str, data: bytes
    ) -> Future[list[Model]]:
        """Asynchronous version of ``upload_models``.

        :param instance: Instance.
        :type instance: Instance
        :param format: Model format (see loader class variables).
        :type format: str
        :param data: Model binary data.
        :type data: bytes
        :return: Future to monitor the task.
        :rtype: Future[list[Model]]
        """
        params = _get_params(self, format=format)
        task = instance.task("upload-model", params, data)
        return Future(task, _get_models)


class MeshLoader(Loader):
    """Mesh loader.

    Main supported formats are OBJ, PLY, STL and OFF.

    Format support can be queried using ``get_loaders(instance)``.
    """

    OBJ: ClassVar[str] = "obj"
    PLY: ClassVar[str] = "ply"
    STL: ClassVar[str] = "stl"
    OFF: ClassVar[str] = "off"

    @classmethod
    @property
    def name(cls) -> str:
        """Get the loader name.

        :return: Loader name.
        :rtype: str
        """
        return "mesh"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {}


class MhdVolumeLoader(Loader):
    """MDH volume loader.

    Supports .mhd file format.
    """

    MHD: ClassVar[str] = "mhd"

    @classmethod
    @property
    def name(cls) -> str:
        """Get the loader name.

        :return: Loader name.
        :rtype: str
        """
        return "mhd-volume"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {}


class VolumeDataType(Enum):
    """Supported volume voxel data types."""

    UNSIGNED_CHAR = "unsigned_char"
    SHORT = "short"
    UNSIGNED_SHORT = "unsigned_short"
    HALF_FLOAT = "half_float"
    FLOAT = "float"
    DOUBLE = "double"


@dataclass
class RawVolumeLoader(Loader):
    """Raw volume loader.

    Supports .raw file format.

    :param dimensions: Volume 3D grid dimensions.
    :type dimensions: Vector3.
    :param spacing: World-space dimensions of a voxel.
    :type size: Vector3.
    :param data_type: Type of data to interpret the volume bytes.
    :type data_type: str.
    """

    RAW: ClassVar[str] = "raw"

    dimensions: Vector3
    spacing: Vector3
    data_type: VolumeDataType

    @classmethod
    @property
    def name(cls) -> str:
        """Get the loader name.

        :return: Loader name.
        :rtype: str
        """
        return "raw-volume"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {
            "dimensions": list(self.dimensions),
            "spacing": list(self.spacing),
            "data_type": self.data_type.value,
        }


def _deserialize_loader(message: dict[str, Any]) -> LoaderInfo:
    return LoaderInfo(
        name=message["name"],
        extensions=message["extensions"],
        schema=deserialize_schema(message["input_parameters_schema"]),
        binary=message.get("binary", False),
        plugin=message.get("plugin", ""),
    )


def _get_params(
    loader: Loader, path: str | None = None, format: str | None = None
) -> dict[str, Any]:
    params = {
        "loader_name": loader.name,
        "loader_properties": loader.get_properties(),
    }
    if path is not None:
        params["path"] = path
    if format is not None:
        params["type"] = format
    return params


def _get_models(reply: JsonRpcReply) -> list[Model]:
    return [deserialize_model(model) for model in reply.result]



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from collections.abc import Callable, Iterator
from typing import Generic, TypeVar

from .jsonrpc import JsonRpcFuture, JsonRpcProgress, JsonRpcReply

T = TypeVar("T")


class Future(Generic[T]):
    """JSON-RPC future wrapper to include the reply processing.

    Add a new method ``wait_for_result`` to retreive the result of the reply
    processing.
    """

    def __init__(
        self, future: JsonRpcFuture, reply_handler: Callable[[JsonRpcReply], T]
    ) -> None:
        self._future = future
        self._reply_handler = reply_handler

    def __iter__(self) -> Iterator[JsonRpcProgress]:
        yield from self._future

    def is_ready(self) -> bool:
        return self._future.is_ready()

    def has_progress(self) -> bool:
        return self._future.has_progress()

    def get_progress(self) -> JsonRpcProgress:
        return self._future.get_progress()

    def wait_for_reply(self) -> JsonRpcReply:
        return self._future.wait_for_reply()

    def wait_for_result(self) -> T:
        reply = self.wait_for_reply()
        return self._reply_handler(reply)

    def poll(self, block: bool) -> None:
        self._future.poll(block)

    def cancel(self) -> None:
        self._future.cancel()



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

import math
from collections.abc import Callable, Iterable
from typing import TypeVar

T = TypeVar("T", bound="Vector")
U = TypeVar("U", int, float)


class Vector(tuple[U, ...]):
    """Generic vector base class.

    Provide either componentwise (with another vector) and scalar operators (
    apply the scalar to all components).

    Examples of arithmetic operations with vector = [1, 1, 1]:

        vector + vector # 2 2 2
        vector * 3 # 3 3 3
        vector / 2 # 0.5 0.5 0.5
        2 / vector # 2 2 2
        vector = Vector3.unpack(i * i for i in range(3)) # 1 4 9
        vector = vector.normalized # vector.norm ~ 1
        vector.square_norm # 3
        vector.norm # sqrt(3)

    See below for supported operators.
    """

    @classmethod
    def unpack(cls: type[T], components: Iterable[U]) -> T:
        return cls(*components)

    def __new__(cls: type[T], *components: U) -> T:
        return super().__new__(cls, components)

    def __str__(self) -> str:
        name = type(self).__name__
        values = ", ".join(str(i) for i in self)
        return f"{name}({values})"

    def __repr__(self) -> str:
        return self.__str__()

    def __neg__(self: T) -> T:
        return self.unpack(-i for i in self)

    def __pos__(self: T) -> T:
        return self.unpack(+i for i in self)

    def __abs__(self: T) -> T:
        return self.unpack(abs(i) for i in self)

    def __add__(self: T, other: T) -> T:
        return self.unpack(i + j for i, j in zip(self, other))

    def __sub__(self: T, other: T) -> T:
        return self.unpack(i - j for i, j in zip(self, other))

    def __mul__(self: T, value: int | float | T) -> T:
        return self.__unpack(value, lambda x, y: x * y)

    def __rmul__(self: T, value: int | float | T) -> T:
        return self.__unpack(value, lambda x, y: y * x)

    def __truediv__(self: T, value: int | float | T) -> T:
        return self.__unpack(value, lambda x, y: x / y)

    def __rtruediv__(self: T, value: int | float | T) -> T:
        return self.__unpack(value, lambda x, y: y / x)

    def __floordiv__(self: T, value: int | float | T) -> T:
        return self.__unpack(value, lambda x, y: x // y)

    def __rfloordiv__(self: T, value: int | float | T) -> T:
        return self.__unpack(value, lambda x, y: y // x)

    def __mod__(self: T, value: int | float | T) -> T:
        return self.__unpack(value, lambda x, y: x % y)

    def __rmod__(self: T, value: int | float | T) -> T:
        return self.__unpack(value, lambda x, y: y % x)

    def __pow__(self: T, value: int | float | T) -> T:
        return self.__unpack(value, lambda x, y: x**y)

    def __rpow__(self: T, value: int | float | T) -> T:
        return self.__unpack(value, lambda x, y: y**x)

    @property
    def square_norm(self) -> float:
        return sum(i * i for i in self)

    @property
    def norm(self) -> float:
        return math.sqrt(self.square_norm)

    @property
    def normalized(self: T) -> T:
        return self / self.norm

    def dot(self: T, other: T) -> float:
        return sum(i * j for i, j in zip(self, other))

    def __unpack(
        self: T, value: int | float | T, operation: Callable[[float, float], float]
    ) -> T:
        if isinstance(value, (int, float)):
            return self.unpack(operation(i, value) for i in self)
        return self.unpack(operation(i, j) for i, j in zip(self, value))


class Vector2(Vector[float]):
    """2D vector with XY components.

    Provides dot product in addition to Vector operators.

    :param x: X component.
    :type x: float
    :param y: Y component.
    :type y: float
    """

    def __new__(cls, x: float = 0.0, y: float = 0.0) -> Vector2:
        return super().__new__(cls, x, y)

    @property
    def x(self) -> float:
        return self[0]

    @property
    def y(self) -> float:
        return self[1]


class Vector3(Vector2):
    """3D vector with XYZ components.

    Provides dot and cross product in addition to Vector operators.

    :param x: X component.
    :type x: float
    :param y: Y component.
    :type y: float
    :param z: Z component.
    :type z: float
    """

    @staticmethod
    def from_vector2(value: Vector2, z: float = 0.0) -> Vector3:
        return Vector3(value.x, value.y, z)

    @classmethod
    @property
    def zero(cls) -> Vector3:
        return Vector3(0.0, 0.0, 0.0)

    @classmethod
    @property
    def one(cls) -> Vector3:
        return Vector3(1.0, 1.0, 1.0)

    def __new__(cls, x: float = 0.0, y: float = 0.0, z: float = 0.0) -> Vector3:
        return Vector[float].__new__(cls, x, y, z)

    @property
    def x(self) -> float:
        return self[0]

    @property
    def y(self) -> float:
        return self[1]

    @property
    def z(self) -> float:
        return self[2]

    @property
    def xy(self) -> Vector2:
        return Vector2(self.x, self.y)

    @property
    def xz(self) -> Vector2:
        return Vector2(self.x, self.z)

    @property
    def yz(self) -> Vector2:
        return Vector2(self.y, self.z)

    def cross(self, other: Vector3) -> Vector3:
        return Vector3(
            self.y * other.z - self.z * other.y,
            self.z * other.x - self.x * other.z,
            self.x * other.y - self.y * other.x,
        )


def componentwise_min(values: list[Vector3]) -> Vector3:
    """Return minimum of each component among values.

    If values is empty, zero is returned.

    :param values: List of vectors.
    :type values: list[Vector3]
    :return: Min value for each component.
    :rtype: Vector3
    """
    if not values:
        return Vector3.zero
    return Vector3(
        min(value.x for value in values),
        min(value.y for value in values),
        min(value.z for value in values),
    )


def componentwise_max(values: list[Vector3]) -> Vector3:
    """Return maximum of each component among values.

    If values is empty, zero is returned.

    :param values: List of vectors.
    :type values: list[Vector3]
    :return: Max value for each component.
    :rtype: Vector3
    """
    if not values:
        return Vector3.zero
    return Vector3(
        max(value.x for value in values),
        max(value.y for value in values),
        max(value.z for value in values),
    )


class Axis:
    """Helper class to store the principal axes of Brayns coordinate system."""

    @classmethod
    @property
    def x(cls) -> Vector3:
        return Vector3(1.0, 0.0, 0.0)

    @classmethod
    @property
    def right(cls) -> Vector3:
        return cls.x

    @classmethod
    @property
    def left(cls) -> Vector3:
        return -cls.right

    @classmethod
    @property
    def y(cls) -> Vector3:
        return Vector3(0.0, 1.0, 0.0)

    @classmethod
    @property
    def up(cls) -> Vector3:
        return cls.y

    @classmethod
    @property
    def down(cls) -> Vector3:
        return -cls.up

    @classmethod
    @property
    def z(cls) -> Vector3:
        return Vector3(0.0, 0.0, 1.0)

    @classmethod
    @property
    def front(cls) -> Vector3:
        return cls.z

    @classmethod
    @property
    def back(cls) -> Vector3:
        return -cls.front



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from collections.abc import Callable, Iterator

from .json_rpc_progress import JsonRpcProgress
from .json_rpc_reply import JsonRpcReply
from .json_rpc_task import JsonRpcTask


class JsonRpcFuture:
    """Future used to monitor a running JSON-RPC request.

    Can be iterated to yield ``JsonRpcProgress`` if the request send any.

    Use ``wait_for_reply`` to wait and get the reply or raise the error.

    Use ``poll`` to refresh the state of the request.

    Use ``cancel`` to send a message to cancel the request.

    Future ready and progress state can also be queried.
    """

    @staticmethod
    def from_reply(reply: JsonRpcReply) -> JsonRpcFuture:
        return JsonRpcFuture(
            task=JsonRpcTask.from_reply(reply),
            cancel=lambda: None,
            poll=lambda _: None,
        )

    def __init__(
        self,
        task: JsonRpcTask,
        cancel: Callable[[], None],
        poll: Callable[[bool], None],
    ) -> None:
        self._task = task
        self._cancel = cancel
        self._poll = poll

    def __iter__(self) -> Iterator[JsonRpcProgress]:
        while True:
            while self.has_progress():
                yield self.get_progress()
            if self.is_ready():
                return
            self.poll(block=True)

    def is_ready(self) -> bool:
        return self._task.is_ready()

    def has_progress(self) -> bool:
        return self._task.has_progress()

    def get_progress(self) -> JsonRpcProgress:
        return self._task.get_progress()

    def wait_for_reply(self) -> JsonRpcReply:
        for _ in self:
            pass
        return self._task.get_reply()

    def poll(self, block: bool) -> None:
        self._poll(block)

    def cancel(self) -> None:
        self._cancel()



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Brayns utilities.

This subpackage contains all functionalities that are commonly used inside and
outside the main package and are not directly related to the Web API.

It includes helpers for algebra, imaging or coloring and the base exception.
"""

from .bounds import Bounds, deserialize_bounds, merge_bounds, serialize_bounds
from .color import Color3, Color4, parse_hex_color
from .error import Error
from .image import ImageFormat, Resolution, parse_image_format
from .json_schema import JsonSchema, JsonType, deserialize_schema, serialize_schema
from .plane_equation import PlaneEquation
from .quaternion import Quaternion
from .rotation import CameraRotation, ModelRotation, Rotation, euler
from .transform import Transform, deserialize_transform, serialize_transform
from .vector import Axis, Vector, Vector2, Vector3, componentwise_max, componentwise_min

__all__ = [
    "Axis",
    "Bounds",
    "CameraRotation",
    "Color3",
    "Color4",
    "componentwise_max",
    "componentwise_min",
    "deserialize_bounds",
    "deserialize_schema",
    "deserialize_transform",
    "Error",
    "euler",
    "ImageFormat",
    "JsonSchema",
    "JsonType",
    "merge_bounds",
    "ModelRotation",
    "parse_hex_color",
    "parse_image_format",
    "PlaneEquation",
    "Quaternion",
    "Resolution",
    "Rotation",
    "serialize_bounds",
    "serialize_schema",
    "serialize_transform",
    "Transform",
    "Vector",
    "Vector2",
    "Vector3",
]



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

import threading
from collections import deque

from websockets.client import WebSocketClientProtocol
from websockets.exceptions import ConnectionClosed

from .errors import ConnectionClosedError, ProtocolError


class AsyncWebSocket:
    def __init__(self, websocket: WebSocketClientProtocol) -> None:
        self._websocket = websocket
        self._queue: deque[bytes | str] = deque(maxlen=100)
        self._condition = threading.Condition()
        self._error: Exception | None = None

    @property
    def closed(self) -> bool:
        return self._websocket.closed

    def poll(self, block: bool) -> bytes | str | None:
        with self._condition:
            data = self._get_data()
            if data is not None:
                return data
            if not block:
                return None
            self._condition.wait()
            return self._get_data()

    async def close(self) -> None:
        await self._websocket.close()

    async def send(self, data: bytes | str) -> None:
        try:
            await self._websocket.send(data)
        except ConnectionClosed as e:
            raise ConnectionClosedError(str(e))

    async def run(self) -> None:
        while self._error is None:
            try:
                data = await self._websocket.recv()
            except ConnectionClosed as e:
                self._set_error(ConnectionClosedError(str(e)))
                return
            except Exception as e:
                self._set_error(ProtocolError(str(e)))
                return
            self._set_data(data)

    def _get_data(self) -> bytes | str | None:
        if self._error is not None:
            raise self._error
        if not self._queue:
            return None
        return self._queue.popleft()

    def _set_data(self, data: bytes | str) -> None:
        with self._condition:
            self._queue.append(data)
            self._condition.notify_all()

    def _set_error(self, e: Exception) -> None:
        with self._condition:
            self._error = e
            self._condition.notify_all()



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging
import sys


class Logger(logging.Logger):
    """Brayns default logger.

    Log to stdout with simple formatting (name, level, message).

    Can be used to change only the log level without creating a new logger.

    :param level: Log level from logging module, defaults to WARN.
    :type level: int, optional
    """

    def __init__(self, level: int = logging.WARN) -> None:
        super().__init__("Brayns", level)
        handler = logging.StreamHandler(sys.stdout)
        format = "[%(name)s][%(levelname)s] %(message)s"
        formatter = logging.Formatter(format)
        handler.setFormatter(formatter)
        self.addHandler(handler)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from dataclasses import dataclass
from enum import Enum
from typing import Any

from brayns.network import Instance


class TimeUnit(Enum):
    """Simulation time unit.

    Simulation time is always in milliseconds but this class is here to avoid
    making this assumption.

    :param MILLISECOND: milliseconds ('ms').
    """

    MILLISECOND = "ms"

    @property
    def seconds(self) -> float:
        """Convert the time unit to seconds.

        :return: Time unit in seconds.
        :rtype: float
        """
        return 1 / self.per_second

    @property
    def per_second(self) -> float:
        """Get the number of time unit per second.

        :return: Time unit count in one second.
        :rtype: float
        """
        return 1000.0


@dataclass
class Simulation:
    """Simulation state.

    :param start_frame: Index of the first frame of the simulation.
    :type start_frame: int
    :param end_frame: Index of the last frame of the simulation.
    :type end_frame: int
    :param current_frame: Index of the current frame of the simulation.
    :type current_frame: int
    :param delta_time: Delta time in ``time_unit`` between two frames.
    :type delta_time: float
    :param time_unit: Time unit, always milliseconds.
    :type time_unit: TimeUnit
    """

    start_frame: int
    end_frame: int
    current_frame: int
    delta_time: float
    time_unit: TimeUnit

    @property
    def frame_count(self) -> int:
        """Number of frames in the simulation.

        :return: Frame count.
        :rtype: int
        """
        return self.end_frame - self.start_frame + 1

    @property
    def duration(self) -> float:
        """Simulation duration in ``time_unit``.

        :return: Duration.
        :rtype: float
        """
        return (self.end_frame - self.start_frame) * self.delta_time

    @property
    def fps(self) -> float:
        """Simulation FPS.

        :return: FPS.
        :rtype: float
        """
        return 1 / self.delta_time * self.time_unit.per_second

    def clamp(self, frame: int) -> int:
        """Clamp given frame index inside simulation limits.

        :param frame: Frame index.
        :type frame: int
        :return: Clamped frame index.
        :rtype: int
        """
        return min(max(frame, self.start_frame), self.end_frame)

    def get_frame(self, time: float) -> int:
        """Convert timestamp in ``time_unit`` to frame index.

        Result is not clamped to simulation limits.

        :param time: Timestep in ``time_unit``.
        :type time: float
        :return: Frame index.
        :rtype: int
        """
        return round(time / self.delta_time)

    def get_time(self, frame: int) -> float:
        """Convert frame index to a timestep in ``time_unit``.

        Result is not clamped to simulation limits.

        :param frame: Frame index.
        :type frame: int
        :return: Timestep in ``time_unit``.
        :rtype: float
        """
        return frame * self.delta_time


def get_simulation(instance: Instance) -> Simulation:
    """Retreive the current simulation state of an instance.

    :param instance: Instance.
    :type instance: Instance
    :return: Current simulation state.
    :rtype: Simulation
    """
    result = instance.request("get-simulation-parameters")
    return _deserialize_simulation(result)


def set_simulation_frame(instance: Instance, index: int) -> None:
    """Set the current simulation frame of an instance.

    :param instance: Instance.
    :type instance: Instance
    :param index: Index of the current frame.
    :type index: int
    """
    params = {"current": index}
    instance.request("set-simulation-parameters", params)


def enable_simulation(instance: Instance, model_id: int, enabled: bool) -> None:
    """Enable the simulation coloring for the given model.

    If enabled, the colors of the model are the one of the simulation at the
    current frame (see ``get_simulation`` and ``set_simulation_frame``).

    If disabled, the colors of the model are set manually or by default.

    :param instance: Instance.
    :type instance: Instance
    :param model_id: Model ID.
    :type model_id: int
    :param enabled: Simulation coloring enabled for given model.
    :type enabled: bool
    """
    params = {
        "model_id": model_id,
        "enabled": enabled,
    }
    instance.request("enable-simulation", params)


def _deserialize_simulation(message: dict[str, Any]) -> Simulation:
    return Simulation(
        start_frame=message["start_frame"],
        end_frame=message["end_frame"],
        current_frame=message["current"],
        delta_time=message["dt"],
        time_unit=TimeUnit(message["unit"]),
    )



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from typing import Any, Protocol, TypeVar

from .jsonrpc import JsonRpcFuture, JsonRpcReply, JsonRpcRequest

T = TypeVar("T", bound="Instance")


class Instance(Protocol):
    """Object used to interact with a connected instance of the backend.

    Contains the connection state with a backend instance and provide methods to
    interact with it.

    An instance is a running process of braynsService executable that runs a
    websocket server and performs the rendering tasks.

    The backend instance can be running on the local machine or a remote one.

    Once finished using an instance, user must call disconnect() on it or use
    its context manager to do it automatically when it goes out of scope.

    The recommended way of doing this is the context manager because it is safe
    regarding exceptions (ie with instance: ...).

    If a websocket communication error occurs, a WebSocketError is raised.

    If a JSON-RPC error message is replied, a RequestError is raised.

    For non-blocking requests, the future is iterable and yield RequestProgess
    objects. Call wait_for_result() to wait for the result.
    """

    def __enter__(self: T) -> T:
        """Context manager enter, just return self."""
        return self

    def __exit__(self, *_) -> None:
        """Context manager exit, close the connection."""
        self.disconnect()

    @property
    def connected(self) -> bool:
        """Check if the instance is connected.

        :return: Connection state.
        :rtype: bool
        """
        return True

    def disconnect(self) -> None:
        """Disconnect instance (mandatory once done with it)."""
        pass

    def request(self, method: str, params: Any = None) -> Any:
        """Send a request to the instance and wait its result.

        Generate automatically the JSON-RPC ID using integers.

        :param method: JSON-RPC method (entrypoint name).
        :type method: str
        :param params: JSON-RPC params, defaults to None
        :type params: Any, optional
        :return: JSON-RPC result.
        :rtype: Any
        """
        reply = self.execute(method, params)
        return reply.result

    def execute(
        self, method: str, params: Any = None, binary: bytes = b""
    ) -> JsonRpcReply:
        """Extended version of request to accept and return binary data.

        :param method: JSON-RPC method.
        :type method: str
        :param params: JSON-RPC params, defaults to None
        :type params: Any, optional
        :param binary: Binary data to send with the request, defaults to b''
        :type binary: bytes, optional
        :return: JSON-RPC result and binary data received with the reply.
        :rtype: tuple[Any, bytes]
        """
        task = self.task(method, params, binary)
        return task.wait_for_reply()

    def task(
        self, method: str, params: Any = None, binary: bytes = b""
    ) -> JsonRpcFuture:
        """Send a request to the instance in a non-blocking way.

        Generate automatically the JSON-RPC ID using integers.

        Use the future returned to wait for progresses or results.

        :param method: JSON-RPC method.
        :type method: str
        :param params: JSON-RPC params, defaults to None
        :type params: Any, optional
        :return: Future object to monitor the request.
        :rtype: RequestFuture
        """
        id = 0
        while self.is_running(id):
            id += 1
        request = JsonRpcRequest(id, method, params, binary)
        return self.send(request)

    def is_running(self, id: int | str) -> bool:
        """Check if the request with the given ID is running.

        :param id: JSON-RPC request ID.
        :type id: int | str
        :return: Running state.
        :rtype: bool
        """
        return False

    def send(self, request: JsonRpcRequest) -> JsonRpcFuture:
        """Send a request in a non blocking way.

        This method is the most basic one, it doesn't generate any ID and can be
        used asynchronously (doesn't block until result is received).

        :param request: JSON-RPC request to send.
        :type request: JsonRpcRequest
        :return: Future object to monitor the request.
        :rtype: RequestFuture
        """
        raise NotImplementedError()

    def poll(self, block: bool) -> None:
        """Check if messages are received from the instance.

        When a message is received, it will update all existing futures.

        The futures call this method in blocking mode when iterated.

        :param block: Wait until a message is received if True.
        :type block: bool
        """
        pass

    def cancel(self, id: int | str) -> None:
        """Cancel request with given ID.

        Send a cancellation request to the instance for a given task.

        :param id: JSON-RPC request ID.
        :type id: int | str
        :raises RequestError: Task is not running.
        """
        pass



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from typing import Protocol, TypeVar

T = TypeVar("T", bound="WebSocket")


class WebSocket(Protocol):
    def __enter__(self: T) -> T:
        return self

    def __exit__(self, *_) -> None:
        self.close()

    @property
    def closed(self) -> bool:
        raise NotImplementedError()

    def close(self) -> None:
        raise NotImplementedError()

    def poll(self, block: bool) -> None:
        raise NotImplementedError()

    def send_binary(self, data: bytes) -> None:
        raise NotImplementedError()

    def send_text(self, data: str) -> None:
        raise NotImplementedError()



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

import logging

from .json_rpc_dispatcher import JsonRpcDispatcher
from .json_rpc_error import JsonRpcError
from .json_rpc_handler import JsonRpcHandler
from .json_rpc_task import JsonRpcTask
from .json_rpc_tasks import JsonRpcTasks


class JsonRpcManager:
    def __init__(self, logger: logging.Logger) -> None:
        self._logger = logger
        self._tasks = JsonRpcTasks()
        listener = JsonRpcHandler(self._tasks, self._logger)
        self._dispatcher = JsonRpcDispatcher(listener)

    def is_running(self, id: int | str) -> bool:
        return id in self._tasks

    def clear(self) -> None:
        self._logger.debug("Clear all JSON-RPC tasks.")
        error = JsonRpcError.general("Disconnection from client side")
        self._tasks.add_error(error)

    def create_task(self, id: int | str) -> JsonRpcTask:
        self._logger.debug("Create JSON-RPC task with ID %s.", id)
        return self._tasks.create_task(id)

    def process_binary(self, data: bytes) -> None:
        self._logger.debug("Processing JSON-RPC binary message: %s.", data)
        self._dispatcher.dispatch_binary(data)

    def process_text(self, data: str) -> None:
        self._logger.debug("Processing JSON-RPC text message: %s.", data)
        self._dispatcher.dispatch_text(data)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from brayns.network import Instance
from brayns.utils import Axis, Rotation, Vector3


@dataclass
class View:
    """Represent a viewpoint with position and target.

    Up and direction doesn't need to be perpendicular. However, having direction
    equal to zero or colinear with up will make an invalid view.

    :param position: Observation position.
    :type position: Vector3
    :param target: Target position.
    :type target: Vector3
    :param up: Up vector, defaults to Axis.up.
    :type up: Vector3, optional
    """

    position: Vector3
    target: Vector3
    up: Vector3 = Axis.up

    @classmethod
    @property
    def front(cls) -> View:
        """Front view with X right, Y up and Z front.

        :return: Front view.
        :rtype: View
        """
        return View(
            position=Vector3.zero,
            target=Axis.back,
        )

    @property
    def vector(self) -> Vector3:
        """Get view vector (target - position).

        :return: View vector.
        :rtype: Vector3
        """
        return self.target - self.position

    @property
    def direction(self) -> Vector3:
        """Get normalized view vector.

        :return: View vector normalized.
        :rtype: Vector3
        """
        return self.vector.normalized

    @property
    def right(self) -> Vector3:
        """Get right direction relative to view (direction x up).

        :return: Right direction normalized.
        :rtype: Vector3
        """
        return self.direction.cross(self.up).normalized

    @property
    def real_up(self) -> Vector3:
        """Get effective up direction (right x direction).

        Equal to up if up is perpendicular to direction and normalized.

        :return: Real up direction normalized.
        :rtype: Vector3
        """
        return self.right.cross(self.direction)

    @property
    def distance(self) -> float:
        """Get the distance between the observator and the target.

        :return: View vector norm.
        :rtype: float
        """
        return self.vector.norm

    @distance.setter
    def distance(self, value: float) -> None:
        """Move the observer position to be at given distance from target.

        View direction remains unchanged.

        :param value: New observer distance.
        :type value: float
        """
        self.position = self.target - value * self.direction

    @property
    def orientation(self) -> Rotation:
        """Get orientation compared to the front view.

        :return: Rotation from front view to self.
        :rtype: Rotation
        """
        return self.get_orientation(View.front)

    def get_orientation(self, reference: View) -> Rotation:
        """Get orientation from given reference.

        The rotation obtained goes from reference.direction and
        reference.real_up to self.direction and self.real_up.

        :return: Rotation from given view to self.
        :rtype: Rotation
        """
        first = Rotation.between(reference.direction, self.direction)
        up = first.apply(reference.real_up)
        second = Rotation.between(up, self.real_up)
        return first.then(second)

    def translate(self, translation: Vector3) -> View:
        """Translate position and target in given direction.

        :param translation: Translation to apply on position and target.
        :type translation: Vector3
        :return: New translated view.
        :rtype: View
        """
        return View(
            position=self.position + translation,
            target=self.target + translation,
            up=self.up,
        )

    def rotate_around_target(self, rotation: Rotation) -> View:
        """Rotate observation position around target.

        :param rotation: Rotation to apply on observer.
        :type rotation: Rotation
        :return: New rotated view.
        :rtype: View
        """
        return View(
            position=rotation.apply(self.position, center=self.target),
            target=self.target,
            up=rotation.apply(self.up),
        )


def get_camera_view(instance: Instance) -> View:
    """Get the current camera view of an instance.

    :param instance: Instance.
    :type instance: Instance
    :return: Camera view.
    :rtype: View
    """
    result = instance.request("get-camera-view")
    return deserialize_view(result)


def set_camera_view(instance: Instance, view: View) -> None:
    """Set the current camera view of an instance.

    :param instance: Instance.
    :type instance: Instance
    :param view: New camera view.
    :type view: View
    """
    params = serialize_view(view)
    instance.request("set-camera-view", params)


def deserialize_view(obj: dict[str, Any]) -> View:
    return View(
        position=Vector3(*obj["position"]),
        target=Vector3(*obj["target"]),
        up=Vector3(*obj["up"]),
    )


def serialize_view(view: View) -> dict[str, Any]:
    return {
        "position": list(view.position),
        "target": list(view.target),
        "up": list(view.up),
    }



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from enum import Enum
from pathlib import Path

from .vector import Vector


class ImageFormat(Enum):
    """Enumeration of all supported image format.

    Values are the image format file extension without the dot.

    :param JPEG: JPEG format, smaller but lose quality.
    :param PNG: PNG format, bigger but lossless.
    """

    JPEG = "jpg"
    PNG = "png"


def parse_image_format(filename: str | Path) -> ImageFormat:
    """Parse the image format from a file path using its extension.

    Supports both string and Path input.

    :param filename: Image file path.
    :type filename: str | Path
    :return: Image format.
    :rtype: ImageFormat
    """
    path = Path(filename) if isinstance(filename, str) else filename
    extension = path.suffix[1:]
    extension = extension.lower()
    extension = extension.strip()
    return ImageFormat(extension)


class Resolution(Vector[int]):
    """Image resolution.

    Resolution is a Vector of integers and can be handled as such.

    :param width: Image width in pixels.
    :type width: int
    :param height: Image height in pixels.
    :type height: int
    """

    @classmethod
    @property
    def full_hd(cls) -> Resolution:
        """Create a full HD (1920x1080) resolution.

        :return: Full HD resolution.
        :rtype: Resolution
        """
        return cls(1920, 1080)

    @classmethod
    @property
    def ultra_hd(cls) -> Resolution:
        """Create a 4K (3840x2160) resolution.

        :return: Ultra HD resolution.
        :rtype: Resolution
        """
        return 2 * cls.full_hd

    @classmethod
    @property
    def production(cls) -> Resolution:
        """Create a production (15360x8640) resolution.

        :return: Production resolution.
        :rtype: Resolution
        """
        return 8 * cls.full_hd

    def __new__(cls, width: int, height: int) -> Resolution:
        if width <= 0 or height <= 0:
            raise ValueError(f"Invalid resolution: {width}x{height}")
        return super().__new__(cls, width, height)

    @property
    def width(self) -> int:
        return self[0]

    @property
    def height(self) -> int:
        return self[1]

    @property
    def aspect_ratio(self) -> float:
        """Get aspect ratio.

        :return: Width / height.
        :rtype: float
        """
        return self.width / self.height



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum

from .process import Process


@dataclass
class SslServerContext:
    """Server SSL context.

    Can be used to start a braynsService instance with SSL enabled. Optional
    parameters such as the server private key, certificate and trusted CAs can
    be specified here.
    """

    private_key_file: str | None = None
    private_key_passphrase: str | None = None
    certificate_file: str | None = None
    ca_location: str | None = None


class LogLevel(Enum):
    """Server (braynsService backend) log level."""

    TRACE = "trace"
    DEBUG = "debug"
    INFO = "info"
    WARN = "warn"
    ERROR = "error"
    CRITICAL = "critical"
    OFF = "off"


class Plugin(Enum):
    """All built-in plugins for braynsService.

    Plugins are loaded when the backend instance is started and cannot be
    changed afterward.

    The value is the name of the plugin dynamic library (.so).
    """

    CIRCUIT_EXPLORER = "braynsCircuitExplorer"
    ATLAS_EXPLORER = "braynsAtlasExplorer"
    CYLINDRIC_CAMERA = "braynsCylindricCamera"
    DTI = "braynsDTI"
    MOLECULE_EXPLORER = "braynsMoleculeExplorer"

    @classmethod
    @property
    def all(cls) -> list[str]:
        """Shortcut to get all the plugin names."""
        return [plugin.value for plugin in Plugin]


@dataclass
class Service:
    """Class used to start a braynsService subprocess.

    Use a braynsService executable to start a subprocess. By default it looks
    for a 'braynsService' binary in the PATH but it can be changed.

    URI is the websocket server URI (ip:port). Use 0.0.0.0 as wildcard to allow
    connections from any machine.

    SSL server settings can be specified using optional certificate, key, CA and
    password.

    The backend log level can also be specified using ``log_level``.

    Custom environment variables can also be set for the subprocess, for example
    to override the PATH and load specific libraries.

    :param uri: Service URI with format 'host:port'.
    :type uri: str
    :param ssl_context: SSL context if secure, defaults to None.
    :type ssl_context: SslServerContext | None, optional
    :param max_clients: Max simultaneous connections authorized, defaults to 1.
    :type max_clients: int, optional
    :param log_level: Process log level, defaults to LogLevel.WARN.
    :type log_level: LogLevel, optional
    :param plugins: Plugins to load, defaults to all built-in plugins.
    :type plugins: list[str], optional
    :param executable: braynsService executable, defaults to 'braynService'.
    :type executable: str, optional
    :param env: Subprocess environment variables, default to empty.
    :type env: dict[str, str], optional
    """

    uri: str
    ssl_context: SslServerContext | None = None
    max_clients: int = 1
    log_level: LogLevel = LogLevel.WARN
    plugins: list[str] = field(default_factory=lambda: Plugin.all)
    executable: str = "braynsService"
    env: dict[str, str] = field(default_factory=dict)

    def get_command_line(self) -> list[str]:
        """Build the command line to start braynsService.

        :return: Command line arguments.
        :rtype: list[str]
        """
        args = [
            self.executable,
            "--uri",
            self.uri,
            "--max-clients",
            str(self.max_clients),
            "--log-level",
            self.log_level.value,
            *_get_plugins_args(self.plugins),
        ]
        if self.ssl_context is not None:
            args.append("--secure")
            args.append("true")
            ssl_args = _get_ssl_args(self.ssl_context)
            args.extend(ssl_args)
        return args

    def start(self) -> Process:
        """Start a new process for a braynsService backend.

        Return the process which runs the service.

        :return: Service process.
        :rtype: Process
        """
        args = self.get_command_line()
        return Process(args, self.env)


def _get_plugins_args(plugins: list[str]) -> list[str]:
    return [arg for plugin in plugins for arg in ("--plugin", plugin)]


def _get_ssl_args(context: SslServerContext) -> list[str]:
    args = list[str]()
    if context.private_key_file is not None:
        args.append("--private-key-file")
        args.append(context.private_key_file)
    if context.private_key_passphrase is not None:
        args.append("--private-key-passphrase")
        args.append(context.private_key_passphrase)
    if context.certificate_file is not None:
        args.append("--certificate-file")
        args.append(context.certificate_file)
    if context.ca_location is not None:
        args.append("--ca-location")
        args.append(context.ca_location)
    return args



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from collections.abc import Callable
from dataclasses import dataclass

from brayns.utils import Axis, Bounds, CameraRotation, Rotation, Vector3

from .camera import Camera
from .projection import PerspectiveProjection, Projection
from .view import View


@dataclass
class CameraController:
    """Helper to create a camera to look at a given target.

    Aspect ratio is usually the one of the resolution of the final image.

    Camera translation and rotation can be used to adjust the camera manually.
    They are applied in this order relatively to the front view (use X to move
    right, Y to move up and Z to move further from the target).

    A new projection is created for each camera. It can be configured using the
    ``projection`` parameter, which is a callable returning a new projection.

    A camera using the current settings can be built at any time using the
    ``camera`` property.

    :param target: Camera target bounds.
    :type target: Bounds
    :param aspect_ratio: Viewport aspect ratio (width / height), defaults to 1.
    :type aspect_ratio: float
    :param translation: Camera translation (before rotation), defaults to zero.
    :type translation: Vector3, optional
    :param rotation: Camera rotation, defaults to identity.
    :type rotation: Rotation, optional
    :param projection: Projection factory, defaults to PerspectiveProjection.
    :type projection: Projection, optional
    """

    target: Bounds
    aspect_ratio: float = 1.0
    translation: Vector3 = Vector3.zero
    rotation: Rotation = CameraRotation.front
    projection: Callable[[], Projection] = PerspectiveProjection

    @property
    def camera(self) -> Camera:
        """Create a new camera using current controller settings.

        :return: Camera focusing on controller target.
        :rtype: Camera
        """
        projection = self.projection()
        target = _get_apparent_target(self.target, self.rotation.inverse)
        view = _get_front_view(target, self.aspect_ratio, projection)
        view = view.translate(self.translation)
        view = view.rotate_around_target(self.rotation)
        return Camera(view, projection)


def _get_apparent_target(target: Bounds, rotation: Rotation) -> Bounds:
    return target.rotate(rotation, target.center)


def _get_front_view(
    target: Bounds, aspect_ratio: float, projection: Projection
) -> View:
    distance = _get_camera_distance(target.size, aspect_ratio, projection)
    center = target.center
    return View(
        position=center + distance * Axis.front,
        target=center,
        up=Axis.up,
    )


def _get_camera_distance(
    target_size: Vector3, aspect_ratio: float, projection: Projection
) -> float:
    width, height, depth = target_size
    height = max(height, width / aspect_ratio)
    return projection.look_at(height) + depth / 2



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, TypeVar

from brayns.network import Instance
from brayns.utils import Color3

T = TypeVar("T", bound="Material")


@dataclass
class Material(ABC):
    """Base class for all material types.

    Material are applied on model to change their aspect (but not their color).
    """

    @classmethod
    @property
    @abstractmethod
    def name(cls) -> str:
        """Get the material name.

        :return: Material name
        :rtype: str
        """
        pass

    @abstractmethod
    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        pass

    @abstractmethod
    def update_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        pass

    @classmethod
    def from_properties(cls: type[T], message: dict[str, Any]) -> T:
        """Low level API to deserialize from JSON."""
        material = cls()
        material.update_properties(message)
        return material


@dataclass
class CarPaintMaterial(Material):
    """Car paint material.

    :param flake_density: Flake density (0-1), defaults to no flakes.
    :type flake_density: float, optional
    """

    flake_density: float = 0.0

    @classmethod
    @property
    def name(cls) -> str:
        """Get the material name.

        :return: Material name
        :rtype: str
        """
        return "carpaint"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {"flake_density": self.flake_density}

    def update_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        self.flake_density = message["flake_density"]


@dataclass
class EmissiveMaterial(Material):
    """Emissive material.

    :param intensity: Light emission intensity, defaults to 1.
    :type intensity: float, optional
    :param color: Light color, defaults to white.
    :type color: Color3, optional
    """

    intensity: float = 1.0
    color: Color3 = Color3.white

    @classmethod
    @property
    def name(cls) -> str:
        """Get the material name.

        :return: Material name
        :rtype: str
        """
        return "emissive"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {
            "intensity": self.intensity,
            "color": list(self.color),
        }

    def update_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        self.intensity = message["intensity"]
        self.color = Color3.unpack(message["color"])


@dataclass
class GlassMaterial(Material):
    """Glass material.

    :param refraction_index: Refraction index, defaults to 1.5.
    :type refraction_index: float, optional
    """

    refraction_index: float = 1.5

    @classmethod
    @property
    def name(cls) -> str:
        """Get the material name.

        :return: Material name
        :rtype: str
        """
        return "glass"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {"index_of_refraction": self.refraction_index}

    def update_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        self.refraction_index = message["index_of_refraction"]


@dataclass
class GhostMaterial(Material):
    """Ghost material."""

    @classmethod
    @property
    def name(cls) -> str:
        """Get the material name.

        :return: Material name
        :rtype: str
        """
        return "ghost"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {}

    def update_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        pass


@dataclass
class MatteMaterial(Material):
    """Matte material.

    :param opacity: Opacity (0-1), defaults to fully opaque.
    :type opacity: float, optional
    """

    opacity: float = 1.0

    @classmethod
    @property
    def name(cls) -> str:
        """Get the material name.

        :return: Material name
        :rtype: str
        """
        return "matte"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {"opacity": self.opacity}

    def update_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        self.opacity = message["opacity"]


@dataclass
class MetalMaterial(Material):
    """Metal material.

    :param roughness: Roughness of the metal, defaults to 1.
    :type roughness: float, optional
    """

    roughness: float = 1.0

    @classmethod
    @property
    def name(cls) -> str:
        """Get the material name.

        :return: Material name
        :rtype: str
        """
        return "metal"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {"roughness": self.roughness}

    def update_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        self.roughness = message["roughness"]


@dataclass
class PhongMaterial(Material):
    """Phong material used by default, works with all renderers.

    :param opacity: Opacity (0-1), defaults to fully opaque.
    :type opacity: float, optional
    """

    opacity: float = 1.0

    @classmethod
    @property
    def name(cls) -> str:
        """Get the material name.

        :return: Material name
        :rtype: str
        """
        return "phong"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {"opacity": self.opacity}

    def update_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        self.opacity = message["opacity"]


@dataclass
class PlasticMaterial(Material):
    """Plastic material.

    :param opacity: Opacity (0-1), defaults to fully opaque.
    :type opacity: float, optional
    """

    opacity: float = 1.0

    @classmethod
    @property
    def name(cls) -> str:
        """Get the material name.

        :return: Material name
        :rtype: str
        """
        return "plastic"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {"opacity": self.opacity}

    def update_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        self.opacity = message["opacity"]


@dataclass
class PrincipledMaterial(Material):
    """Principled material.

    :param edge_color: Surface edge tint for metallic surfaces.
    :type edge_color: float, optional
    :param metallic: Alpha parameter between dielectric and metallic.
    :type metallic: float, optional
    :param diffuse: Diffuse reflection weight.
    :type diffuse: float, optional
    :param specular: Specular reflection/transmission weight.
    :type specular: float, optional
    :param ior: Dielectric index of refraction.
    :type ior: float, optional
    :param transmission: Specular transmission weight.
    :type transmission: float, optional
    :param transmission_color: Transmission attenuation color.
    :type transmission_color: Color3, optional
    :param transmission_depth: Surface distance for pure transmission color.
    :type transmission_depth: float, optional
    :param roughness: Diffuse and specular reflection roughness.
    :type roughness: float, optional
    :param anisotropy: Specular anisotropy reflection weight.
    :type anisotropy: float, optional
    :param anisotropy_rotation: Rotation of the specular anisotropy reflection.
    :type anisotropy_rotation: float, optional
    :param thin: Specified wether the object is solid or thin (hollow).
    :type thin: bool, optional
    :param thickness: Thickness of the object if thin = True.
    :type thickness: float, optional
    :param back_light: Weight of reflection and transmission if thin = True.
    :type back_light: float, optional
    :param coat: Clear coat weight.
    :type coat: float, optional
    :param coat_ior: Clear coat index of refraction.
    :type coat_ior: float, optional
    :param coat_color: Clear coat color.
    :type coat_color: Color3, optional
    :param coat_thickness: Clear coat thickness.
    :type coat_thickness: float, optional
    :param coat_roughness: Clear coat diffuse/specular reflection roughness.
    :type coat_roughness: float, optional
    :param sheen: Sheen effect weight.
    :type sheen: float, optional
    :param sheen_color: Sheen color.
    :type sheen_color: Color3, optional
    :param sheen_tint: Alpha from white to sheen color for sheen effect.
    :type sheen_tint: float, optional
    :param sheen_roughness: Sheen diffuse/specular reflection roughness.
    :type sheen_roughness: float, optional
    """

    edge_color: Color3 = Color3(1, 1, 1)
    metallic: float = 0
    diffuse: float = 1
    specular: float = 1
    ior: float = 1
    transmission: float = 0
    transmission_color: Color3 = Color3(1, 1, 1)
    transmission_depth: float = 1
    roughness: float = 1
    anisotropy: float = 0
    anisotropy_rotation: float = 0
    thin: bool = False
    thickness: float = 1
    back_light: float = 0
    coat: float = 0
    coat_ior: float = 1.5
    coat_color: Color3 = Color3(1, 1, 1)
    coat_thickness: float = 1
    coat_roughness: float = 0
    sheen: float = 0
    sheen_color: Color3 = Color3(1, 1, 1)
    sheen_tint: float = 0
    sheen_roughness: float = 0.2

    @classmethod
    @property
    def name(cls) -> str:
        """Get the material name.

        :return: Material name
        :rtype: str
        """
        return "principled"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {
            "edge_color": list(self.edge_color),
            "metallic": self.metallic,
            "diffuse": self.diffuse,
            "specular": self.specular,
            "ior": self.ior,
            "transmission": self.transmission,
            "transmission_color": list(self.transmission_color),
            "transmission_depth": self.transmission_depth,
            "roughness": self.roughness,
            "anisotropy": self.anisotropy,
            "anisotropy_rotation": self.anisotropy_rotation,
            "thin": self.thin,
            "thickness": self.thickness,
            "back_light": self.back_light,
            "coat": self.coat,
            "coat_ior": self.coat_ior,
            "coat_color": list(self.coat_color),
            "coat_thickness": self.coat_thickness,
            "coat_roughness": self.coat_roughness,
            "sheen": self.sheen,
            "sheen_color": list(self.sheen_color),
            "sheen_tint": self.sheen_tint,
            "sheen_roughness": self.sheen_roughness,
        }

    def update_properties(self, message: dict[str, Any]) -> None:
        """Low level API to deserialize from JSON."""
        self.edge_color = Color3.unpack(message["edge_color"])
        self.metallic = message["metallic"]
        self.diffuse = message["diffuse"]
        self.specular = message["specular"]
        self.ior = message["ior"]
        self.transmission = message["transmission"]
        self.transmission_color = Color3.unpack(message["transmission_color"])
        self.transmission_depth = message["transmission_depth"]
        self.roughness = message["roughness"]
        self.anisotropy = message["anisotropy"]
        self.anisotropy_rotation = message["anisotropy_rotation"]
        self.thin = message["thin"]
        self.thickness = message["thickness"]
        self.back_light = message["back_light"]
        self.coat = message["coat"]
        self.coat_ior = message["coat_ior"]
        self.coat_color = Color3.unpack(message["coat_color"])
        self.coat_thickness = message["coat_thickness"]
        self.coat_roughness = message["coat_roughness"]
        self.sheen = message["sheen"]
        self.sheen_color = Color3.unpack(message["sheen_color"])
        self.sheen_tint = message["sheen_tint"]
        self.sheen_roughness = message["sheen_roughness"]


def get_material_name(instance: Instance, model_id: int) -> str:
    """Get the name of the material applied on the given model.

    Can be used to check that a given material is applied on a model:

    :param instance: Instance.
    :type instance: Instance
    :param model_id: ID of the model to check.
    :type model_id: int
    :return: Material name.
    :rtype: str
    """
    params = {"id": model_id}
    return instance.request("get-material-type", params)


def get_material(instance: Instance, model_id: int, material_type: type[T]) -> T:
    """Get the material applied on the given model.

    ``material_type`` must be the current model material type.

    :param instance: Instance.
    :type instance: Instance
    :param model_id: ID of the model.
    :type model_id: int
    :param material_type: Material type.
    :type material_type: type[T]
    :return: Material applied on ``model``.
    :rtype: T
    """
    name = material_type.name
    params = {"id": model_id}
    result = instance.request(f"get-material-{name}", params)
    return material_type.from_properties(result)


def set_material(instance: Instance, model_id: int, material: Material) -> None:
    """Apply the given material to the given model.

    :param instance: Instance.
    :type instance: Instance
    :param model_id: Model ID.
    :type model_id: int
    :param material: Material to apply on model.
    :type material: Material
    """
    name = material.name
    params = {
        "model_id": model_id,
        "material": material.get_properties(),
    }
    instance.request(f"set-material-{name}", params)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from brayns.network import Future, Instance, JsonRpcFuture, JsonRpcReply
from brayns.utils import ImageFormat, Resolution, parse_image_format

from .camera import Camera
from .renderer import Renderer
from .view import serialize_view


@dataclass
class ImageMetadata:
    """Metadata information that will be embeded into the image.

    :param title: Image title.
    :type title: str
    :param description: Image description.
    :type description: str
    :param where_used: Event for which the image was created.
    :type where_used: str
    :param keywords: List of keywords that describe the image content.
    :type keywords: list[str]
    """

    title: str
    description: str
    where_used: str
    keywords: list[str]


@dataclass
class Snapshot:
    """Helper class to take a snapshot of an instance with custom settings.

    Snapshots create a temporary context in the instance to avoid changing
    current instance settings.

    For None parameters, the current values of the instance are used.

    :param resolution: Image resolution, defaults to None.
    :type resolution: Resolution | None, optional
    :param camera: Camera used to render, defaults to None.
    :type camera: Camera | None, optional
    :param renderer: Renderer used to render, defaults to None.
    :type renderer: Renderer | None, optional
    :param frame: Simulation index, defaults to None.
    :type frame: int | None, optional
    :param jpeg_quality: JPEG quality if format is JPEG, defaults to 100%.
    :type jpeg_quality: int, optional
    :param metadata: Metadata information to embed into the image.
    :type metadata: ImageMetadata | None, optional
    """

    resolution: Resolution | None = None
    camera: Camera | None = None
    renderer: Renderer | None = None
    frame: int | None = None
    jpeg_quality: int = 100
    metadata: ImageMetadata | None = None

    def save(self, instance: Instance, path: str) -> None:
        """Download and save the snapshot locally under given file.

        Path is on the local machine (running current script).

        :param instance: Instance.
        :type instance: Instance
        :param path: Output file.
        :type path: str
        """
        task = self.save_task(instance, path)
        task.wait_for_result()

    def save_task(self, instance: Instance, path: str) -> Future[None]:
        """Asynchronous version of ``save``.

        :param instance: Instance.
        :type instance: Instance
        :param path: Output file.
        :type path: str
        :return: Future to monitor the task.
        :rtype: Future[None]
        """
        format = parse_image_format(path)
        task = self._download(instance, format)
        return Future(task, lambda reply: _save_color_buffer(reply, path))

    def save_remotely(self, instance: Instance, path: str) -> None:
        """Save the snapshot remotely under given file.

        Path is on the remote machine (running instance backend).

        :param instance: Instance.
        :type instance: Instance
        :param path: Output file.
        :type path: str
        """
        task = self.save_remotely_task(instance, path)
        task.wait_for_result()

    def save_remotely_task(self, instance: Instance, path: str) -> Future[None]:
        """Asynchronous version of ``save_remotely``.

        :param instance: Instance.
        :type instance: Instance
        :param path: Output file.
        :type path: str
        :return: Future to monitor the task.
        :rtype: Future[None]
        """
        format = parse_image_format(path)
        params = _serialize_snapshot(self, format, path)
        task = _task(instance, params)
        return Future(task, lambda _: None)

    def download(
        self, instance: Instance, format: ImageFormat = ImageFormat.PNG
    ) -> bytes:
        """Download the rendered image as bytes at given format.

        :param instance: Instance.
        :type instance: Instance
        :param path: Output file.
        :type path: str
        :return: Image data.
        :rtype: bytes
        """
        task = self.download_task(instance, format)
        return task.wait_for_result()

    def download_task(
        self, instance: Instance, format: ImageFormat = ImageFormat.PNG
    ) -> Future[bytes]:
        """Asynchronous version of ``download``.

        :param instance: Instance.
        :type instance: Instance
        :param path: Output file.
        :type path: str
        :return: Future to monitor the task.
        :rtype: Future[bytes]
        """
        task = self._download(instance, format)
        return Future(task, _get_color_buffer)

    def _download(self, instance: Instance, format: ImageFormat) -> JsonRpcFuture:
        params = _serialize_snapshot(self, format)
        return _task(instance, params)


def _serialize_snapshot(
    snapshot: Snapshot, format: ImageFormat, path: str | None = None
) -> dict[str, Any]:
    message: dict[str, Any] = {
        "image_settings": _serialize_image_settings(snapshot, format),
    }
    if path is not None:
        message["file_path"] = path
    if snapshot.camera is not None:
        camera = snapshot.camera
        message["camera_view"] = serialize_view(camera.view)
        message["camera"] = camera.projection.get_properties_with_name()
    if snapshot.renderer is not None:
        message["renderer"] = snapshot.renderer.get_properties_with_name()
    if snapshot.frame is not None:
        message["simulation_frame"] = snapshot.frame
    if snapshot.metadata is not None:
        message["metadata"] = _serialize_image_metadata(snapshot.metadata)
    return message


def _serialize_image_settings(
    snapshot: Snapshot, format: ImageFormat
) -> dict[str, Any]:
    message: dict[str, Any] = {
        "format": format.value,
    }
    if format is ImageFormat.JPEG:
        message["quality"] = snapshot.jpeg_quality
    if snapshot.resolution is not None:
        message["size"] = list(snapshot.resolution)
    return message


def _serialize_image_metadata(metadata: ImageMetadata) -> dict[str, Any]:
    return {
        "title": metadata.title,
        "description": metadata.description,
        "where_used": metadata.where_used,
        "keywords": metadata.keywords,
    }


def _task(instance: Instance, params: dict[str, Any]) -> JsonRpcFuture:
    return instance.task("snapshot", params)


def _get_color_buffer(reply: JsonRpcReply) -> bytes:
    result, data = reply.result, reply.binary
    buffer = result["color_buffer"]
    offset = buffer["offset"]
    size = buffer["size"]
    return data[offset:size]


def _save_color_buffer(reply: JsonRpcReply, path: str) -> None:
    data = _get_color_buffer(reply)
    with open(path, "wb") as file:
        file.write(data)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from dataclasses import dataclass
from typing import Any

from brayns.network import Instance
from brayns.utils import Resolution


@dataclass
class Application:
    """Store the application parameters of an instance.

    :param plugins: List of plugins loaded in the instance.
    :type plugins: list[str]
    :param resolution: Framebuffer resolution.
    :type resolution: Resolution
    """

    plugins: list[str]
    resolution: Resolution


def get_application(instance: Instance) -> Application:
    """Retreive the application parameters from an instance.

    :param instance: Instance.
    :type instance: Instance
    :return: Current application parameters.
    :rtype: Application
    """
    result = instance.request("get-application-parameters")
    return _deserialize_application(result)


def set_resolution(instance: Instance, resolution: Resolution) -> None:
    """Update the framebuffer resolution of the given instance.

    :param instance: Instance.
    :type instance: Instance
    :param resolution: Framebuffer resolution.
    :type resolution: Resolution
    """
    params = {"viewport": list(resolution)}
    instance.request("set-application-parameters", params)


def stop(instance: Instance) -> None:
    """Stop a running instance.

    :param instance: Instance.
    :type instance: Instance
    """
    instance.request("quit")


def _deserialize_application(message: dict[str, Any]) -> Application:
    return Application(
        plugins=message["plugins"],
        resolution=Resolution(*message["viewport"]),
    )



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Brayns core functionalities.

Provides high-level objects to wrap the raw JSON-RPC messages sent and received
with an instance of Brayns service.

Include access to camera, renderer, models, materials, snapshots, etc...
"""

from .application import Application, get_application, set_resolution, stop
from .camera import Camera, get_camera, set_camera
from .camera_controller import CameraController
from .color_ramp import ColorRamp, ValueRange, get_color_ramp, set_color_ramp
from .coloring import (
    ColorMethod,
    color_model,
    get_color_methods,
    get_color_values,
    set_model_color,
)
from .entrypoint import Entrypoint, get_entrypoint, get_entrypoints, get_methods
from .framebuffer import (
    Framebuffer,
    ProgressiveFramebuffer,
    StaticFramebuffer,
    set_framebuffer,
)
from .gbuffer_exporter import GBufferChannel, GBufferExporter
from .geometry import (
    BoundedPlane,
    Box,
    Capsule,
    Geometry,
    Plane,
    Sphere,
    add_clipping_geometries,
    add_geometries,
    clear_clipping_geometries,
)
from .image import Image, ImageInfo
from .light import (
    AmbientLight,
    DirectionalLight,
    Light,
    QuadLight,
    add_light,
    clear_lights,
)
from .loader import (
    Loader,
    LoaderInfo,
    MeshLoader,
    MhdVolumeLoader,
    RawVolumeLoader,
    VolumeDataType,
    get_loaders,
)
from .material import (
    CarPaintMaterial,
    EmissiveMaterial,
    GhostMaterial,
    GlassMaterial,
    Material,
    MatteMaterial,
    MetalMaterial,
    PhongMaterial,
    PlasticMaterial,
    PrincipledMaterial,
    get_material,
    get_material_name,
    set_material,
)
from .model import (
    Model,
    Scene,
    clear_models,
    clear_renderables,
    deserialize_model,
    get_bounds,
    get_model,
    get_models,
    get_scene,
    instantiate_model,
    remove_models,
    update_model,
)
from .opacity_curve import ControlPoint, OpacityCurve
from .pick import PickResult, pick
from .projection import (
    Fovy,
    OrthographicProjection,
    PerspectiveProjection,
    Projection,
    get_camera_name,
    get_camera_projection,
    set_camera_projection,
)
from .renderer import (
    InteractiveRenderer,
    ProductionRenderer,
    Renderer,
    get_renderer,
    get_renderer_name,
    set_renderer,
)
from .simulation import (
    Simulation,
    TimeUnit,
    enable_simulation,
    get_simulation,
    set_simulation_frame,
)
from .snapshot import ImageMetadata, Snapshot
from .version import Version, VersionError, check_version, get_version
from .view import (
    View,
    deserialize_view,
    get_camera_view,
    serialize_view,
    set_camera_view,
)

__all__ = [
    "add_clipping_geometries",
    "add_geometries",
    "add_light",
    "AmbientLight",
    "Application",
    "BoundedPlane",
    "Box",
    "Camera",
    "CameraController",
    "Capsule",
    "CarPaintMaterial",
    "check_version",
    "clear_clipping_geometries",
    "clear_lights",
    "clear_models",
    "clear_renderables",
    "color_model",
    "ColorMethod",
    "ColorRamp",
    "ControlPoint",
    "deserialize_model",
    "deserialize_view",
    "DirectionalLight",
    "EmissiveMaterial",
    "enable_simulation",
    "Entrypoint",
    "Fovy",
    "Framebuffer",
    "GBufferChannel",
    "GBufferExporter",
    "Geometry",
    "get_application",
    "get_bounds",
    "get_camera_name",
    "get_camera_projection",
    "get_camera_view",
    "get_camera",
    "get_color_methods",
    "get_color_ramp",
    "get_color_values",
    "get_entrypoint",
    "get_entrypoints",
    "get_loaders",
    "get_material_name",
    "get_material",
    "get_methods",
    "get_model",
    "get_models",
    "get_renderer_name",
    "get_renderer",
    "get_scene",
    "get_simulation",
    "get_version",
    "GhostMaterial",
    "GlassMaterial",
    "Image",
    "ImageInfo",
    "ImageMetadata",
    "instantiate_model",
    "InteractiveRenderer",
    "Light",
    "Loader",
    "LoaderInfo",
    "Material",
    "MatteMaterial",
    "MeshLoader",
    "MetalMaterial",
    "MhdVolumeLoader",
    "Model",
    "OpacityCurve",
    "OrthographicProjection",
    "PerspectiveProjection",
    "PhongMaterial",
    "pick",
    "PickResult",
    "Plane",
    "PlasticMaterial",
    "PrincipledMaterial",
    "ProductionRenderer",
    "ProgressiveFramebuffer",
    "Projection",
    "QuadLight",
    "RawVolumeLoader",
    "remove_models",
    "Renderer",
    "Scene",
    "serialize_view",
    "set_camera_projection",
    "set_camera_view",
    "set_camera",
    "set_color_ramp",
    "set_framebuffer",
    "set_material",
    "set_model_color",
    "set_renderer",
    "set_resolution",
    "set_simulation_frame",
    "Simulation",
    "Snapshot",
    "Sphere",
    "StaticFramebuffer",
    "stop",
    "TimeUnit",
    "update_model",
    "ValueRange",
    "Version",
    "VersionError",
    "View",
    "VolumeDataType",
]



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from collections.abc import Iterator
from dataclasses import dataclass

from brayns.core import Simulation


@dataclass
class MovieFrames:
    """Helper class to generate frame indices.

    Use the movie specification (FPS, slow motion, frame range) to generate
    the indices of the simulation frames which needs to be rendered.

    Slowing factor is compared to real time (2 = twice slower).

    Start and end frames are clamped to the simulation limits and are threated
    like Python indices (i.e. -1 is simulation.end_frame).

    The index step is computed as simulation_fps / fps / slowing_factor. If it
    is smaller than 1, then some frames will be duplicated to match the target
    FPS.

    :param fps: Movie FPS.
    :type fps: float
    :param slowing_factor: Slowing factor.
    :type slowing_factor: float
    :param start_frame: First frame of the movie.
    :type start_frame: int
    :param end_frame: Last frame of the movie.
    :type end_frame: int
    """

    fps: float = 25.0
    slowing_factor: float = 1.0
    start_frame: int = 0
    end_frame: int = -1

    def get_indices(self, simulation: Simulation) -> list[int]:
        start = _get_frame(simulation, self.start_frame)
        end = _get_frame(simulation, self.end_frame)
        step = _get_step(simulation, self)
        return list(_yield_indices(start, end, step))


def _get_frame(simulation: Simulation, frame: int) -> int:
    if frame < 0:
        frame += simulation.end_frame + 1
    return simulation.clamp(frame)


def _get_step(simulation: Simulation, frames: MovieFrames) -> float:
    return simulation.fps / frames.fps / frames.slowing_factor


def _yield_indices(start: int, end: int, step: float) -> Iterator[int]:
    current = float(start)
    while True:
        if current > end:
            return
        yield round(current)
        current += step



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, TypeVar

from brayns.network import Instance
from brayns.utils import Bounds, Color4, PlaneEquation, Vector3, serialize_bounds

from .model import Model, deserialize_model


@dataclass
class Geometry(ABC):
    """Base class of all geometry types."""

    @classmethod
    @property
    @abstractmethod
    def name(cls) -> str:
        """Low level API to serialize to JSON."""
        pass

    @abstractmethod
    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        pass


@dataclass
class BoundedPlane(Geometry):
    """Axis-aligned bounded plane.

    :param equation: Plane equation coefficients.
    :type equation: PlaneEquation
    :param bounds: Axis aligned bounds to limit the plane.
    :type bounds: Bounds
    """

    equation: PlaneEquation
    bounds: Bounds

    def __post_init__(self) -> None:
        if min(self.bounds.size) <= 0:
            raise ValueError("All bound dimensions must be greater than 0")

    @classmethod
    @property
    def name(cls) -> str:
        """Low level API to serialize to JSON."""
        return "bounded-planes"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {
            "coefficients": list(self.equation),
            "bounds": serialize_bounds(self.bounds),
        }


@dataclass
class Box(Geometry):
    """3D box.

    :param min: Minimum XYZ.
    :type min: Vector3
    :param max: Maximum XYZ.
    :type max: Vector3
    """

    min: Vector3
    max: Vector3

    @classmethod
    @property
    def name(cls) -> str:
        """Low level API to serialize to JSON."""
        return "boxes"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {
            "min": list(self.min),
            "max": list(self.max),
        }


@dataclass
class Capsule(Geometry):
    """Capsule (cylinder with slope and half spheres at extremities).

    :param start_point: Cylinder base XYZ.
    :type start_point: Vector3
    :param start_radius: Cylinder and sphere base radius.
    :type start_radius: float
    :param end_point: Cylinder end XYZ.
    :type end_point: Vector3
    :param end_radius: Cylinder and sphere end radius.
    :type end_radius: float
    """

    start_point: Vector3
    start_radius: float
    end_point: Vector3
    end_radius: float

    @classmethod
    @property
    def name(cls) -> str:
        """Low level API to serialize to JSON."""
        return "capsules"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {
            "p0": list(self.start_point),
            "r0": self.start_radius,
            "p1": list(self.end_point),
            "r1": self.end_radius,
        }


@dataclass
class Plane(Geometry):
    """Infinite plane.

    :param equation: Plane equation coefficients.
    :type equation: PlaneEquation
    """

    equation: PlaneEquation

    @classmethod
    @property
    def name(cls) -> str:
        """Low level API to serialize to JSON."""
        return "planes"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {
            "coefficients": list(self.equation),
        }


@dataclass
class Sphere(Geometry):
    """Sphere with radius and position.

    :param radius: Radius.
    :type radius: float
    :param center: Center XYZ, defaults to origin.
    :type center: Vector3
    """

    radius: float
    center: Vector3 = Vector3.zero

    @classmethod
    @property
    def name(cls) -> str:
        """Low level API to serialize to JSON."""
        return "spheres"

    def get_properties(self) -> dict[str, Any]:
        """Low level API to serialize to JSON."""
        return {
            "center": list(self.center),
            "radius": self.radius,
        }


T = TypeVar("T", bound="Geometry")


def add_geometries(instance: Instance, geometries: list[tuple[T, Color4]]) -> Model:
    """Create a model from a list of geometries with optional color.

    All geometries must have the same type.

    Model witout geometries are not supported.

    :param instance: Instance.
    :type instance: Instance
    :param geometries: Geometries to add (boxes, capsules, etc...).
    :type geometries: list[T]
    :raises ValueError: List is empty.
    :return: Model created from the geometries.
    :rtype: Model
    """
    if not geometries:
        raise ValueError("Cannot create a model with no geometries")
    method = f"add-{geometries[0][0].name}"
    params = [
        {
            "geometry": geometry.get_properties(),
            "color": list(color),
        }
        for geometry, color in geometries
    ]
    result = instance.request(method, params)
    return deserialize_model(result)


def add_clipping_geometries(
    instance: Instance, geometries: list[T], invert_normals: bool = False
) -> Model:
    """Create a model from a list of clipping geometries.

    All geometries must have the same type.

    Model witout geometries are not supported.

    :param instance: Instance.
    :type instance: Instance
    :param geometries: Clipping geometries to add (boxes, capsules, etc...).
    :type geometries: list[T]
    :param invert_normals: Invert shading normals to switch clipping side.
    :type invert_normals: bool
    :raises ValueError: List is empty.
    :return: Model created from the geometries.
    :rtype: Model
    """
    if not geometries:
        raise ValueError("Cannot create a model with no clipping geometries")
    method = f"add-clipping-{geometries[0].name}"
    params: dict[str, Any] = {
        "primitives": [geometry.get_properties() for geometry in geometries],
        "invert_normals": invert_normals,
    }
    result = instance.request(method, params)
    return deserialize_model(result)


def clear_clipping_geometries(instance: Instance) -> None:
    """Clear all clipping geometries from the given instance.

    :param instance: Instance.
    :type instance: Instance
    """
    instance.request("clear-clipping-geometries")



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from brayns.network import Instance
from brayns.utils import (
    Bounds,
    Transform,
    deserialize_bounds,
    deserialize_transform,
    serialize_transform,
)


@dataclass
class Model:
    """Loaded model.

    All models are loaded without transform (identity) but it doesn't mean that
    their center of mass is at the origin (depends on the source file).

    Model metadata are a str -> str map and depends on the model type.

    :param id: Model ID.
    :type id: int
    :param type: Model type.
    :type type: str
    :param bounds: Model bounding box.
    :type bounds: Bounds
    :param info: Intrinsic nodel information.
    :type info: dict[str, Any]
    :param visible: Check wether the model is rendered or not.
    :type visible: bool
    :param transform: Model transform relative to the origin.
    :type transform: Transform
    """

    id: int
    type: str
    bounds: Bounds
    info: dict[str, Any]
    visible: bool
    transform: Transform


@dataclass
class Scene:
    """Contains all existing models and scene boundaries.

    Scene bounds take models and lights into account.

    :param bounds: Global bounds.
    :type bounds: Bounds
    :param models: List of all existing models.
    :type models: list[Model]
    """

    bounds: Bounds
    models: list[Model]


def get_model(instance: Instance, id: int) -> Model:
    """Retreive the model with given ID from an instance.

    :param instance: Instance.
    :type instance: Instance
    :param id: Model ID.
    :type id: int
    :return: Model object.
    :rtype: Model
    """
    result = instance.request("get-model", {"id": id})
    return deserialize_model(result)


def update_model(
    instance: Instance,
    model_id: int,
    transform: Transform | None = None,
    visible: bool | None = None,
) -> Model:
    """Modify the properties of a given model and return its updated version.

    Parameters left as None will remain in their current state.

    :param instance: Instance.
    :type instance: Instance
    :param model_id: ID of the model to update.
    :type model_id: int
    :param transform: Model transformation, defaults to None
    :type transform: Transform | None, optional
    :param visible: Model visibility, defaults to None
    :type visible: bool | None, optional
    :return: Updated model.
    :rtype: Model
    """
    properties = dict[str, Any]()
    if transform is not None:
        properties["transform"] = serialize_transform(transform)
    if visible is not None:
        properties["is_visible"] = visible
    params = {
        "model_id": model_id,
        "model": properties,
    }
    result = instance.request("update-model", params)
    return deserialize_model(result)


def remove_models(instance: Instance, ids: list[int]) -> None:
    """Remove the given models from an instance.

    :param instance: Instance.
    :type instance: Instance
    :param ids: ID(s) of the models to remove.
    :type ids: list[int]
    """
    params = {"ids": ids}
    instance.request("remove-model", params)


def clear_models(instance: Instance) -> None:
    """Clear all models from the given instance.

    :param instance: Instance.
    :type instance: Instance
    """
    instance.request("clear-models")


def clear_renderables(instance: Instance) -> None:
    """Clear all models that can be rendered from the given instance.

    Renderables exclude lights and clipping geometries.

    :param instance: Instance.
    :type instance: Instance
    """
    instance.request("clear-renderables")


def get_scene(instance: Instance) -> Scene:
    """Retreive all models and the world boundaries from an instance.

    :param instance: Instance.
    :type instance: Instance
    :return: Models and global bounds.
    :rtype: Scene
    """
    result = instance.request("get-scene")
    return _deserialize_scene(result)


def get_bounds(instance: Instance) -> Bounds:
    """Retreive the scene boundary of an instance.

    The scene boundaries are computed from all existing lights and models
    in the given instance.

    :param instance: Instance.
    :type instance: Instance
    :return: Bounds of the current scene.
    :rtype: Bounds
    """
    return get_scene(instance).bounds


def get_models(instance: Instance) -> list[Model]:
    """Retreive all models from an instance.

    :param instance: Instance.
    :type instance: Instance
    :return: List of models.
    :rtype: list[Model]
    """
    return get_scene(instance).models


def instantiate_model(
    instance: Instance, model_id: int, transforms: list[Transform]
) -> list[Model]:
    """Create instances of the given model and return them.

    The amount of new instances is equal to the number of transforms passed.

    :param instance: Instance.
    :type instance: Instance
    :param model_id: ID of the model to instantiate.
    :type model_id: int
    :param transforms: List of new instances transformations.
    :type transforms: list[Transform]
    :return: New instance model.
    :rtype: Model
    """

    if len(transforms) == 0:
        raise ValueError("transforms list cannot be empty")

    params: dict[str, Any] = {
        "model_id": model_id,
        "transforms": [serialize_transform(transform) for transform in transforms],
    }

    models = instance.request("instantiate-model", params)
    return [deserialize_model(model) for model in models]


def deserialize_model(message: dict[str, Any]) -> Model:
    return Model(
        id=message["model_id"],
        type=message["model_type"],
        bounds=deserialize_bounds(message["bounds"]),
        info=message.get("info", {}),
        visible=message["is_visible"],
        transform=deserialize_transform(message["transform"]),
    )


def _deserialize_scene(message: dict[str, Any]) -> Scene:
    return Scene(
        bounds=deserialize_bounds(message["bounds"]),
        models=[deserialize_model(model) for model in message["models"]],
    )



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from dataclasses import dataclass

from brayns.utils import Color3, Color4


@dataclass
class ControlPoint:
    """Control point used for opacity curves.

    Normalized values are simulation values (usually voltages) normalized
    between the min and max of the color ramp value range.

    :param normalized_value: Simulation value, normalized.
    :type normalized_value: float
    :param opacity: Color opacity at simulation value.
    :type opacity: float
    """

    normalized_value: float
    opacity: float

    @classmethod
    @property
    def start(cls) -> ControlPoint:
        """Implicit first control point of the opacity curve [0, 0].

        :return: Control point.
        :rtype: ControlPoint
        """
        return ControlPoint(0.0, 0.0)

    @classmethod
    @property
    def end(cls) -> ControlPoint:
        """Implicit last control point of the opacity curve [1, 1].

        :return: Control point.
        :rtype: ControlPoint
        """
        return ControlPoint(1.0, 1.0)


@dataclass
class OpacityCurve:
    """Opacity curve to generate opacity from a list of color3.

    Use control points to map simulation value (usually voltage) to opacity.

    The simulation values are normalized so it is not necessary to know them.

    If no control points are set at 0 or 1, implicit control points [0, 0] and
    [1, 1] will be used for the computations. Otherwise the user-defined ones
    are used.

    :param control_points: Control points of the curve.
    :type control_points: list[ControlPoint]
    """

    control_points: list[ControlPoint]

    def apply(self, colors: list[Color3]) -> list[Color4]:
        """Create a from colors with opacities computed from the curve.

        Works also with a list of Color4 as lists are covariant in Python.

        Simulation normalized value range is inferred from len(colors).

        :param colors: Input colors.
        :type colors: list[Color3]
        :return: Output colors.
        :rtype: list[Color4]
        """
        result = []
        color_count = len(colors)
        for i in range(color_count):
            normalized_value = _normalize_index(i, color_count)
            opacity = _get_opacity(self.control_points, normalized_value)
            color = Color4.from_color3(colors[i], opacity)
            result.append(color)
        return result


def _get_opacity(control_points: list[ControlPoint], normalized_value: float) -> float:
    lower_index = _find_lower_index(control_points, normalized_value)
    upper_index = lower_index + 1
    lower_control_point = _find(control_points, lower_index)
    upper_control_point = _find(control_points, upper_index)
    lower = lower_control_point.normalized_value
    upper = upper_control_point.normalized_value
    lower_ratio = _get_lower_ratio(normalized_value, lower, upper)
    upper_ratio = 1 - lower_ratio
    lower_opacity = lower_control_point.opacity
    upper_opacity = upper_control_point.opacity
    return lower_opacity * lower_ratio + upper_opacity * upper_ratio


def _find_lower_index(
    control_points: list[ControlPoint], normalized_value: float
) -> int:
    for i, control_point in enumerate(control_points):
        if control_point.normalized_value >= normalized_value:
            return i - 1
    return len(control_points) - 1


def _find(control_points: list[ControlPoint], index: int) -> ControlPoint:
    if index < 0:
        return ControlPoint.start
    if index >= len(control_points):
        return ControlPoint.end
    return control_points[index]


def _normalize_index(index: int, size: int) -> float:
    return index / max(size - 1, 1)


def _get_lower_ratio(value: float, lower: float, upper: float) -> float:
    if lower == upper:
        return 0.0
    gap = upper - value
    size = upper - lower
    return gap / size



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Brayns Python package.

This package provides a high level API to interact with an instance of Brayns
instance through websockets.

The low level JSON-RPC API is also available using the instance directly.
"""

from .core import (
    AmbientLight,
    Application,
    BoundedPlane,
    Box,
    Camera,
    CameraController,
    Capsule,
    CarPaintMaterial,
    ColorMethod,
    ColorRamp,
    ControlPoint,
    DirectionalLight,
    EmissiveMaterial,
    Entrypoint,
    Fovy,
    Framebuffer,
    GBufferChannel,
    GBufferExporter,
    Geometry,
    GhostMaterial,
    GlassMaterial,
    Image,
    ImageInfo,
    ImageMetadata,
    InteractiveRenderer,
    Light,
    Loader,
    LoaderInfo,
    Material,
    MatteMaterial,
    MeshLoader,
    MetalMaterial,
    MhdVolumeLoader,
    Model,
    OpacityCurve,
    OrthographicProjection,
    PerspectiveProjection,
    PhongMaterial,
    PickResult,
    Plane,
    PlasticMaterial,
    PrincipledMaterial,
    ProductionRenderer,
    ProgressiveFramebuffer,
    Projection,
    QuadLight,
    RawVolumeLoader,
    Renderer,
    Scene,
    Simulation,
    Snapshot,
    Sphere,
    StaticFramebuffer,
    TimeUnit,
    ValueRange,
    Version,
    VersionError,
    View,
    VolumeDataType,
    add_clipping_geometries,
    add_geometries,
    add_light,
    check_version,
    clear_clipping_geometries,
    clear_lights,
    clear_models,
    clear_renderables,
    color_model,
    enable_simulation,
    get_application,
    get_bounds,
    get_camera,
    get_camera_name,
    get_camera_projection,
    get_camera_view,
    get_color_methods,
    get_color_ramp,
    get_color_values,
    get_entrypoint,
    get_entrypoints,
    get_loaders,
    get_material,
    get_material_name,
    get_methods,
    get_model,
    get_models,
    get_renderer,
    get_renderer_name,
    get_scene,
    get_simulation,
    get_version,
    instantiate_model,
    pick,
    remove_models,
    set_camera,
    set_camera_projection,
    set_camera_view,
    set_color_ramp,
    set_framebuffer,
    set_material,
    set_model_color,
    set_renderer,
    set_resolution,
    set_simulation_frame,
    stop,
    update_model,
)
from .movie import Movie, MovieError, MovieFrames
from .network import (
    ConnectionClosedError,
    Connector,
    Future,
    Instance,
    InvalidServerCertificateError,
    JsonRpcError,
    JsonRpcFuture,
    JsonRpcProgress,
    JsonRpcReply,
    JsonRpcRequest,
    Logger,
    ProtocolError,
    ServiceUnavailableError,
    SslClientContext,
    WebSocketError,
)
from .plugins import (
    AtlasColumnHighlight,
    AtlasDensity,
    AtlasDistanceType,
    AtlasFlatmapAreas,
    AtlasLayerDistance,
    AtlasOrientationField,
    AtlasShellOutline,
    AtlasUsecase,
    BbpCells,
    BbpLoader,
    BbpReport,
    BbpReportType,
    CellPlacementLoader,
    CircuitColorMethod,
    ColumnNeighbor,
    ColumnPosition,
    CylindricProjection,
    DtiLoader,
    GeometryType,
    Morphology,
    MorphologyColorMethod,
    MorphologyLoader,
    NrrdLoader,
    ProteinColorScheme,
    ProteinLoader,
    SonataEdgePopulation,
    SonataLoader,
    SonataNodePopulation,
    SonataNodes,
    SonataReport,
    SonataReportType,
    VoxelType,
    XyzLoader,
    get_atlas_usecases,
    get_circuit_ids,
    set_circuit_thickness,
)
from .service import (
    LogLevel,
    Manager,
    Plugin,
    Process,
    Service,
    SslServerContext,
    start,
)
from .utils import (
    Axis,
    Bounds,
    CameraRotation,
    Color3,
    Color4,
    Error,
    ImageFormat,
    JsonSchema,
    JsonType,
    ModelRotation,
    PlaneEquation,
    Quaternion,
    Resolution,
    Rotation,
    Transform,
    Vector2,
    Vector3,
    componentwise_max,
    componentwise_min,
    deserialize_schema,
    euler,
    merge_bounds,
    parse_hex_color,
    parse_image_format,
    serialize_schema,
)
from .version import VERSION

__version__ = VERSION
"""Version tag of brayns Python package (major.minor.patch)."""

__all__ = [
    "add_clipping_geometries",
    "add_geometries",
    "add_light",
    "AmbientLight",
    "Application",
    "AtlasColumnHighlight",
    "AtlasDensity",
    "AtlasDistanceType",
    "AtlasFlatmapAreas",
    "AtlasLayerDistance",
    "AtlasOrientationField",
    "AtlasShellOutline",
    "AtlasUsecase",
    "Axis",
    "BbpCells",
    "BbpLoader",
    "BbpReport",
    "BbpReportType",
    "BoundedPlane",
    "Bounds",
    "Box",
    "Camera",
    "CameraController",
    "CameraRotation",
    "Capsule",
    "CarPaintMaterial",
    "CellPlacementLoader",
    "check_version",
    "CircuitColorMethod",
    "clear_clipping_geometries",
    "clear_lights",
    "clear_models",
    "clear_renderables",
    "color_model",
    "Color3",
    "Color4",
    "ColorMethod",
    "ColorRamp",
    "ColumnNeighbor",
    "ColumnPosition",
    "componentwise_max",
    "componentwise_min",
    "ConnectionClosedError",
    "Connector",
    "ControlPoint",
    "CylindricProjection",
    "deserialize_schema",
    "DirectionalLight",
    "DtiLoader",
    "EmissiveMaterial",
    "enable_simulation",
    "Entrypoint",
    "Error",
    "euler",
    "Fovy",
    "Framebuffer",
    "Future",
    "GBufferChannel",
    "GBufferExporter",
    "Geometry",
    "GeometryType",
    "get_application",
    "get_atlas_usecases",
    "get_bounds",
    "get_camera_name",
    "get_camera_projection",
    "get_camera_view",
    "get_camera",
    "get_circuit_ids",
    "get_color_methods",
    "get_color_ramp",
    "get_color_values",
    "get_entrypoint",
    "get_entrypoints",
    "get_loaders",
    "get_material_name",
    "get_material",
    "get_methods",
    "get_model",
    "get_models",
    "get_renderer_name",
    "get_renderer",
    "get_scene",
    "get_simulation",
    "get_version",
    "GhostMaterial",
    "GlassMaterial",
    "Image",
    "ImageFormat",
    "ImageInfo",
    "ImageMetadata",
    "Instance",
    "instantiate_model",
    "InteractiveRenderer",
    "InvalidServerCertificateError",
    "JsonRpcError",
    "JsonRpcFuture",
    "JsonRpcProgress",
    "JsonRpcReply",
    "JsonRpcRequest",
    "JsonSchema",
    "JsonType",
    "Light",
    "Loader",
    "LoaderInfo",
    "Logger",
    "LogLevel",
    "Manager",
    "Material",
    "MatteMaterial",
    "merge_bounds",
    "MeshLoader",
    "MetalMaterial",
    "MhdVolumeLoader",
    "Model",
    "ModelRotation",
    "Morphology",
    "MorphologyColorMethod",
    "MorphologyLoader",
    "Movie",
    "MovieError",
    "MovieFrames",
    "NrrdLoader",
    "OpacityCurve",
    "OrthographicProjection",
    "parse_hex_color",
    "parse_image_format",
    "PerspectiveProjection",
    "PhongMaterial",
    "pick",
    "PickResult",
    "Plane",
    "PlaneEquation",
    "PlasticMaterial",
    "Plugin",
    "PrincipledMaterial",
    "Process",
    "ProductionRenderer",
    "ProgressiveFramebuffer",
    "Projection",
    "ProteinColorScheme",
    "ProteinLoader",
    "ProtocolError",
    "QuadLight",
    "Quaternion",
    "RawVolumeLoader",
    "remove_models",
    "Renderer",
    "Resolution",
    "Rotation",
    "Scene",
    "serialize_schema",
    "Service",
    "ServiceUnavailableError",
    "set_camera_projection",
    "set_camera_view",
    "set_camera",
    "set_circuit_thickness",
    "set_color_ramp",
    "set_framebuffer",
    "set_material",
    "set_model_color",
    "set_renderer",
    "set_resolution",
    "set_simulation_frame",
    "Simulation",
    "Snapshot",
    "SonataEdgePopulation",
    "SonataLoader",
    "SonataNodePopulation",
    "SonataNodes",
    "SonataReport",
    "SonataReportType",
    "Sphere",
    "SslClientContext",
    "SslServerContext",
    "start",
    "StaticFramebuffer",
    "stop",
    "TimeUnit",
    "Transform",
    "update_model",
    "ValueRange",
    "Vector2",
    "Vector3",
    "Version",
    "VersionError",
    "View",
    "VolumeDataType",
    "VoxelType",
    "WebSocketError",
    "XyzLoader",
]



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from dataclasses import dataclass

from brayns.utils import Error


@dataclass
class WebSocketError(Error):
    """Base class of all network exceptions.

    :param reason: Short description of what happened.
    :type reason: str
    """

    reason: str

    def __str__(self) -> str:
        """Display the reason when printed."""
        return self.reason


class ConnectionClosedError(WebSocketError):
    """Raised when doing operation on a disconnected instance.

    Happens after the connection to an instance.
    """


class InvalidServerCertificateError(WebSocketError):
    """Raised when the server certificate is rejected by the client.

    SSL specific error happening during the connection to an instance.
    """


class ProtocolError(WebSocketError):
    """Raised when a websocket protocol error occurs.

    Usually happens during the connection to an instance.
    """


class ServiceUnavailableError(WebSocketError):
    """Raised when braynsService instance is not available at given URI.

    Occurs when the backend instance is not ready or started.

    Can be used to try multiple connections to wait for an instance to be ready.
    """



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
# Responsible Author: Nadir Roman Guerrero <nadir.romanguerrero@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from dataclasses import dataclass
from typing import Any, ClassVar

from brayns.core import Loader


@dataclass
class DtiLoader(Loader):
    """Diffuse Tensor Imaging loader.

    :param streamline_radius: Radius of the streamlines, defaults to 1.
    :type streamline_radius: float, optional
    :param streamline_radius: Radius of the streamlines, defaults to 1.
    :type streamline_radius: float, optional
    """

    JSON: ClassVar[str] = "json"

    streamline_radius: float = 1.0
    spike_decay_time: float = 1.0

    @classmethod
    @property
    def name(cls) -> str:
        return "DTI loader"

    def get_properties(self) -> dict[str, Any]:
        return {
            "radius": self.streamline_radius,
            "spike_decay_time": self.spike_decay_time,
        }



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
# Responsible Author: Nadir Roman Guerrero <nadir.romanguerrero@epfl.ch>
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, ClassVar

from brayns.core import Loader


class GeometryType(Enum):
    """Type of geometry to use when loading a morphology.

    :param ORIGINAL: Use raw geometries dimensions.
    :param SMOOTH: Smooth radius changes between geometries for better visual.
    :param SECTION_SMOOTH: Smooth radius change between whole sections.
    :param CONSTANT_RADII: Apply the same radius to all geometries.
    """

    ORIGINAL = "original"
    SMOOTH = "smooth"
    SECTION_SMOOTH = "section_smooth"
    CONSTANT_RADII = "constant_radii"
    SPHERES = "spheres"


@dataclass
class Morphology:
    """Describe how to load a morphology.

    :param radius_multiplier: Radius scaling, defaults to 1.
    :type radius_multiplier: float, optional
    :param load_soma: Wether to load somas, defaults to True.
    :type load_soma: bool, optional
    :param load_axon: Wether to load axons, defaults to False.
    :type load_axon: bool, optional
    :param load_dendrites: Wether to load dendrites, defaults to False.
    :type load_dendrites: bool, optional
    :param geometry_type: How to load geometries, defaults to smooth.
    :type geometry_type: GeometryType, optional
    :param resampling: Min cos of angle between segments to merge them.
    :type resampling: float
    :param subsampling: Step factor to skip morphology samples.
    :type subsampling: int
    """

    radius_multiplier: float = 1.0
    load_soma: bool = True
    load_axon: bool = False
    load_dendrites: bool = False
    geometry_type: GeometryType = GeometryType.SMOOTH
    resampling: float = 2
    subsampling: int = 1


def serialize_morphology(morphology: Morphology) -> dict[str, Any]:
    return {
        "radius_multiplier": morphology.radius_multiplier,
        "load_soma": morphology.load_soma,
        "load_axon": morphology.load_axon,
        "load_dendrites": morphology.load_dendrites,
        "geometry_type": morphology.geometry_type.value,
        "resampling": morphology.resampling,
        "subsampling": morphology.subsampling,
    }


class MorphologyColorMethod:
    """Available coloring methods for morphologies."""

    MORPHOLOGY_SECTION: ClassVar[str] = "morphology section"
    MORPHOLOGY_SECTION_GEOMETRY: ClassVar[str] = "morphology section geometry"


@dataclass
class MorphologyLoader(Loader):
    """Loader for morphology files.

    :param morphology: How to load the morphologies, default constructed.
    :type morphology: Morphology, optional
    """

    SWC: ClassVar[str] = "swc"
    H5: ClassVar[str] = "h5"
    ASC: ClassVar[str] = "asc"

    morphology: Morphology = field(default_factory=Morphology)

    @classmethod
    @property
    def name(cls) -> str:
        return "Neuron Morphology loader"

    def get_properties(self) -> dict[str, Any]:
        return serialize_morphology(self.morphology)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from .async_web_socket import AsyncWebSocket
from .errors import ConnectionClosedError
from .event_loop import EventLoop
from .web_socket import WebSocket
from .web_socket_listener import WebSocketListener


class WebSocketClient(WebSocket):
    def __init__(
        self, websocket: AsyncWebSocket, loop: EventLoop, listener: WebSocketListener
    ) -> None:
        self._websocket = websocket
        self._loop = loop
        self._listener = listener
        self._task = self._loop.run(self._websocket.run())

    @property
    def closed(self) -> bool:
        return self._websocket.closed and self._loop.closed

    def close(self) -> None:
        if self.closed:
            return
        self._loop.run(self._websocket.close()).result()
        self._task.result()
        self._loop.close()

    def poll(self, block: bool) -> None:
        self._check_closed()
        data = self._websocket.poll(block)
        if data is None:
            return
        if isinstance(data, bytes):
            self._listener.on_binary(data)
            return
        if isinstance(data, str):
            self._listener.on_text(data)
            return
        raise TypeError(f"Invalid packet type {type(data)}")

    def send_binary(self, data: bytes) -> None:
        self._send(data)

    def send_text(self, data: str) -> None:
        self._send(data)

    def _send(self, data: bytes | str) -> None:
        self._check_closed()
        self._loop.run(self._websocket.send(data)).result()

    def _check_closed(self) -> None:
        if self.closed:
            raise ConnectionClosedError("Connection closed")



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Subpackage to connect to a Brayns service instance (backend).

An instance is a wrapper around a websocket connection and a JSON-RPC context.

It provides functionalities to send JSON-RPC requests and receive replies with a
Brayns instance.
"""

from .connector import Connector
from .future import Future
from .instance import Instance
from .jsonrpc import (
    JsonRpcError,
    JsonRpcFuture,
    JsonRpcProgress,
    JsonRpcReply,
    JsonRpcRequest,
)
from .logger import Logger
from .websocket import (
    ConnectionClosedError,
    InvalidServerCertificateError,
    ProtocolError,
    ServiceUnavailableError,
    SslClientContext,
    WebSocketError,
)

__all__ = [
    "ConnectionClosedError",
    "Connector",
    "Future",
    "Instance",
    "InvalidServerCertificateError",
    "JsonRpcError",
    "JsonRpcFuture",
    "JsonRpcProgress",
    "JsonRpcReply",
    "JsonRpcRequest",
    "Logger",
    "ProtocolError",
    "ServiceUnavailableError",
    "SslClientContext",
    "WebSocketError",
]



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass
class JsonRpcError(Exception):
    """Error raised by an instance when a request fails.

    :param id: ID of the request sending the error.
        None for global errors with no requests attached (like invalid JSON).
    :type id: int | str | None
    :param code: Error code.
    :type code: int
    :param message: Error description.
    :type message: str
    :param data: Optional additional error information, defaults to None.
    :type data: Any, optional
    """

    id: int | str | None
    code: int
    message: str
    data: Any = None

    @staticmethod
    def general(message: str) -> JsonRpcError:
        return JsonRpcError(
            id=None,
            code=0,
            message=message,
        )

    def __str__(self) -> str:
        description = self.message
        details = [f"code={self.code}"]
        if self.id is not None:
            details.append(f"request ID={self.id})")
        if self.data is not None:
            details.append(f"data={self.data})")
        detail = ", ".join(details)
        return f"{description} ({detail})"


def deserialize_error(message: dict[str, Any]) -> JsonRpcError:
    error: dict[str, Any] = message["error"]
    return JsonRpcError(
        id=message.get("id"),
        code=error["code"],
        message=error["message"],
        data=error.get("data"),
    )



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

import os
import subprocess
import threading
from collections import deque
from typing import IO, cast


class Process:
    """Process running a braynsService backend.

    The process must be stopped using terminate() once done with it, otherwise
    it will run forever. Use context manager (ie with process: ...) to avoid
    mistakes.

    The last lines of the process logs can be retreived using ``logs``.
    """

    def __init__(self, args: list[str], env: dict[str, str]) -> None:
        self._process = subprocess.Popen(
            args=args,
            env=os.environ | env,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        )
        self._thread = threading.Thread(target=self._poll)
        self._logs = deque[str](maxlen=1000)
        self._lock = threading.RLock()
        self._thread.start()

    def __enter__(self) -> Process:
        """Context manager, just return self."""
        return self

    def __exit__(self, *_) -> None:
        """Context manager, call ``stop()``."""
        self.stop()

    @property
    def alive(self) -> bool:
        """Check if the process is still running."""
        return self._process.poll() is None

    @property
    def logs(self) -> str:
        """Get the last logs (stdout + stderr) from the process."""
        with self._lock:
            return "".join(self._logs)

    def stop(self) -> None:
        """Stop the backend service by terminating the process.

        Must be called through the context manager (i.e. 'with') or manually to
        avoid the process to run forever.
        """
        self._process.terminate()
        self._process.wait()
        stdin = cast(IO[str], self._process.stdin)
        stdin.close()
        stdout = cast(IO[str], self._process.stdout)
        stdout.close()
        self._thread.join()

    def _poll(self) -> None:
        stdout = cast(IO[str], self._process.stdout)
        for line in stdout:
            with self._lock:
                self._logs.append(line)



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, ClassVar

from brayns.core import Loader

from .morphology import Morphology, serialize_morphology


@dataclass
class BbpCells:
    """BlueConfig cells selection.

    Use one of the factory methods to create this object.
    """

    density: float | None = None
    targets: list[str] | None = None
    gids: list[int] | None = None

    @staticmethod
    def all() -> BbpCells:
        """Select all cells from a circuit."""
        return BbpCells.from_density(1.0)

    @staticmethod
    def from_density(density: float) -> BbpCells:
        """Select only a proportion of cells (0 = None, 1 = all)."""
        return BbpCells(density=density)

    @staticmethod
    def from_targets(targets: list[str], density: float = 1.0) -> BbpCells:
        """Select only the cells from a given target.

        The density of cells to load per target can also be specified.
        """
        return BbpCells(density=density, targets=targets)

    @staticmethod
    def from_gids(gids: list[int]) -> BbpCells:
        """Select cells using their GIDs."""
        return BbpCells(gids=gids)


class BbpReportType(Enum):
    """All available BBP report types."""

    NONE = "none"
    SPIKES = "spikes"
    COMPARTMENT = "compartment"


@dataclass
class BbpReport:
    """BlueConfig report selection.

    Use one of the factory methods to create this object.
    """

    type: BbpReportType
    name: str | None = None
    spike_transition_time: float | None = None

    @staticmethod
    def none() -> BbpReport:
        """No reports to load."""
        return BbpReport(
            type=BbpReportType.NONE,
        )

    @staticmethod
    def spikes(spike_transition_time: float = 1.0) -> BbpReport:
        """Spike report with optional transition time.

        time unit of transition depends on the report but it can be used to make
        this transition faster or slower (ie 2 = twice slower).
        """
        return BbpReport(
            type=BbpReportType.SPIKES,
            spike_transition_time=spike_transition_time,
        )

    @staticmethod
    def compartment(name: str) -> BbpReport:
        """Compartment report with given name."""
        return BbpReport(
            type=BbpReportType.COMPARTMENT,
            name=name,
        )


@dataclass
class BbpLoader(Loader):
    """BlueConfig / CircuitConfig loader.

    Use this loader to load circuit files with BBP internal format.

    :param cells: Cells to load, defaults to all.
    :type cells: BbpCells, optional
    :param report: Report to load, defaults to None.
    :type report: BbpReport | None, optional
    :param morphology: Specifies how to load the circuit morphologies.
    :type morphology: Morphology, optional
    :param load_afferent_synapses: Wether to load afferent, defaults to False.
    :type load_afferent_synapses: bool, optional
    :param load_efferent_synapses: Wether to load efferent, defaults to False.
    :type load_efferent_synapses: bool, optional
    """

    BLUE_CONFIG: ClassVar[str] = "BlueConfig"
    CIRCUIT_CONFIG: ClassVar[str] = "CircuitConfig"

    cells: BbpCells = field(default_factory=BbpCells.all)
    report: BbpReport | None = None
    morphology: Morphology = field(default_factory=Morphology)
    load_afferent_synapses: bool = False
    load_efferent_synapses: bool = False

    @classmethod
    @property
    def name(cls) -> str:
        return "BBP loader"

    def get_properties(self) -> dict[str, Any]:
        return _serialize_loader(self)


def _serialize_loader(loader: BbpLoader) -> dict[str, Any]:
    message: dict[str, Any] = {
        "load_afferent_synapses": loader.load_afferent_synapses,
        "load_efferent_synapses": loader.load_efferent_synapses,
        **_serialize_cells(loader.cells),
        "neuron_morphology_parameters": serialize_morphology(loader.morphology),
    }
    if loader.report is not None:
        report = _serialize_report(loader.report)
        message.update(report)
    return message


def _serialize_cells(cells: BbpCells) -> dict[str, Any]:
    message = dict[str, Any]()
    if cells.density is not None:
        message["percentage"] = cells.density
    if cells.targets is not None:
        message["targets"] = cells.targets
    if cells.gids is not None:
        message["gids"] = cells.gids
    return message


def _serialize_report(report: BbpReport) -> dict[str, Any]:
    message: dict[str, Any] = {
        "report_type": report.type.value,
    }
    if report.name is not None:
        message["report_name"] = report.name
    if report.spike_transition_time is not None:
        message["spike_transition_time"] = report.spike_transition_time
    return message



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from collections.abc import Iterator

from .json_rpc_error import JsonRpcError
from .json_rpc_progress import JsonRpcProgress
from .json_rpc_reply import JsonRpcReply
from .json_rpc_task import JsonRpcTask


class JsonRpcTasks:
    def __init__(self) -> None:
        self._tasks: dict[int | str, JsonRpcTask] = {}

    def __len__(self) -> int:
        return len(self._tasks)

    def __iter__(self) -> Iterator[tuple[int | str, JsonRpcTask]]:
        yield from self._tasks.items()

    def __contains__(self, id: int | str) -> bool:
        return id in self._tasks

    def find(self, id: int | str) -> JsonRpcTask | None:
        return self._tasks.get(id)

    def create_task(self, id: int | str) -> JsonRpcTask:
        if id in self._tasks:
            raise ValueError("Request with same ID already running")
        task = JsonRpcTask()
        self._tasks[id] = task
        return task

    def add_reply(self, reply: JsonRpcReply) -> None:
        if reply.id is None:
            raise ValueError("No request ID in reply")
        task = self.find(reply.id)
        if task is None:
            return
        task.set_reply(reply)
        del self._tasks[reply.id]

    def add_error(self, error: JsonRpcError) -> None:
        if error.id is None:
            self._add_general_error(error)
            return
        task = self.find(error.id)
        if task is None:
            return
        task.set_error(error)
        del self._tasks[error.id]

    def add_progress(self, progress: JsonRpcProgress) -> None:
        task = self.find(progress.id)
        if task is None:
            return
        task.add_progress(progress)

    def _add_general_error(self, error: JsonRpcError) -> None:
        for task in self._tasks.values():
            task.set_error(error)
        self._tasks.clear()



# Copyright (c) 2015-2023 EPFL/Blue Brain Project
# All rights reserved. Do not distribute without permission.
#
# Responsible Author: adrien.fleury@epfl.ch
#
# This file is part of Brayns <https://github.com/BlueBrain/Brayns>
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License version 3.0 as published
# by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import annotations

from typing import TypeVar

from .vector import Vector

T = TypeVar("T", bound="Color3")


class Color3(Vector[float]):
    """Color with RGB normalized components.

    Color3 are vectors and can be manipulated as such.

    :param r: Red component 0-1.
    :type r: float
    :param g: Green component 0-1.
    :type g: float
    :param b: Blue component 0-1.
    :type b: float
    """

    @classmethod
    @property
    def black(cls: type[T]) -> T:
        """Create a black color.

        :return: Black color [0, 0, 0].
        :rtype: Color3
        """
        return cls()

    @classmethod
    @property
    def white(cls: type[T]) -> T:
        """Create a white color.

        :return: White color [1, 1, 1].
        :rtype: Color3
        """
        return cls(1.0, 1.0, 1.0)

    @classmethod
    @property
    def red(cls: type[T]) -> T:
        """Create a red color.

        :return: red color [1, 0, 0].
        :rtype: Color3
        """
        return cls(1.0, 0.0, 0.0)

    @classmethod
    @property
    def green(cls: type[T]) -> T:
        """Create a green color.

        :return: green color [0, 1, 0].
        :rtype: Color3
        """
        return cls(0.0, 1.0, 0.0)

    @classmethod
    @property
    def blue(cls: type[T]) -> T:
        """Create a blue color.

        :return: blue color [0, 0, 1].
        :rtype: Color3
        """
        return cls(0.0, 0.0, 1.0)

    @classmethod
    @property
    def bbp_background(cls: type[T]) -> T:
        """Create a color of the BBP standard background (~blue).

        :return: BBP background color [0.004, 0.016, 0.102].
        :rtype: Color3
        """
        return cls(0.004, 0.016, 0.102)

    def __new__(cls: type[T], r: float = 0.0, g: float = 0.0, b: float = 0.0) -> T:
        return super().__new__(cls, r, g, b)

    @property
    def r(self) -> float:
        return self[0]

    @property
    def g(self) -> float:
        return self[1]

    @property
    def b(self) -> float:
        return self[2]


class Color4(Color3):
    """Color with RGBA normalized components.

    Color4 are Color3 and can be manipulated as such but with alpha channel.

    Opaque is alpha = 1 and fully transparent is alpha = 0.

    Color3 standard colors, if used from Color4, will have an alpha of 1.

    :param a: Alpha component, defaults to 1.0.
    :type a: float, optional
    """

    @staticmethod
    def from_color3(color: Color3, alpha: float = 1.0) -> Color4:
        """Helper to build a Color4 from a Color3.

        :param color: Color3 to convert.
        :type color: Color3
        :param alpha: Alpha channel, defaults to 1.0
        :type alpha: float, optional
        :return: Color4 converted.
        :rtype: Color4
        """
        return Color4(color.r, color.g, color.b, alpha)

    def __new__(
        cls, r: float = 0.0, g: float = 0.0, b: float = 0.0, a: float = 1.0
    ) -> Color4:
        return Vector[float].__new__(cls, r, g, b, a)

    @property
    def a(self) -> float:
        return self[3]

    @property
    def transparent(self) -> Color4:
        """Create an identical color but with alpha = 0.

        :return: Color but fully transparent.
        :rtype: Color4
        """
        return Color4(self.r, self.g, self.b, a=0.0)

    @property
    def opaque(self) -> Color4:
        """Create an identical color but with alpha = 1.

        :return: Color but fully opaque.
        :rtype: Color4
        """
        return Color4(self.r, self.g, self.b, a=1.0)

    @property
    def without_alpha(self) -> Color3:
        """Helper class to create a Color3 from a Color4.

        :return: Color3 converted.
        :rtype: Color3
        """
        return Color3(self.r, self.g, self.b)


def parse_hex_color(value: str) -> Color3:
    """Parse an hexadecimal color string to Color3.

    The string can be just digits (0a12f5), prefixed with a hash (#0a12f5) or
    with 0x (0x0a12f5).

    :param value: Color code.
    :type value: str
    :return: Color parsed.
    :rtype: Color3
    """
    value = _sanitize(value)
    return Color3(
        _normalize(value[0:2]), _normalize(value[2:4]), _normalize(value[4:6])
    )


def _sanitize(value: str) -> str:
    size = len(value)
    if size == 6:
        return value
    if size == 7 and value[0] == "#":
        return value[1:]
    if size == 8 and value[:2].lower() == "0x":
        return value[2:]
    raise ValueError(f"Not an hex color {value}")


def _normalize(value: str) -> float:
    return int(value, base=16) / 255
